# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
***
*** Owner: Andrew Ziem
*** Author: Peter Marshall
***
*** References:
*** Windows Internals (Russinovich, Solomon, Ionescu), 6th edition
*** http://windowsitpro.com/systems-management/inside-windows-nt-disk-defragmenting
*** https://technet.microsoft.com/en-us/sysinternals/sdelete.aspx
*** https://blogs.msdn.microsoft.com/jeffrey_wall/2004/09/13/defrag-api-c-wrappers/
*** https://msdn.microsoft.com/en-us/library/windows/desktop/aa364572(v=vs.85).aspx
***
***
*** Algorithm
***   --Phase 1
*** - Check if the file has special characteristics (sparse, encrypted,
***   compressed), determine file system (NTFS or FAT), Windows version.
*** - Read the on-disk locations of the file using defrag API.
*** - If file characteristics don't rule it out, just do a direct write
***   of zero-fill on entire file size and flush to disk.
*** - Read back the on-disk locations of the file using defrag API.
*** - If locations are exactly the same, we are done.
*** - Otherwise, enumerate clusters that did not get overwritten in place
***   ("missed clusters").
***   They are probably still untouched, we need to wipe them.
*** - If it was a special file that wouldn't be wiped by a direct write,
***   we will truncate the file and treat it all as missed clusters.
***
***   --Phase 2
*** - (*) Get volume bitmap of free/allocated clusters using defrag API. 
***   Figure out if checkpoint has made our missed clusters available
***   for use again (this is potentially delayed by a few seconds in NTFS).
*** - If they have not yet been made available, wait 0.1s then repeat
***   previous check (*), up to a limit of 7s in polling.
*** - Figure out if it is better to bridge the extents, wiping more clusters
***   but gaining a performance boost from reduced total cycles and overhead.
*** - Recurse over the extents we need to wipe, breaking them down into
***   smaller extents if necessary.
*** - Write a zero-fill file that will provide enough clusters to
***   completely overwrite each extent in turn.
*** - Iterate over the zero-fill file, moving clusters from our zero file
***   to the missed clusters using defrag API.
*** - If the defrag move operation did not succeed, it was probably because
***   another process has grabbed a cluster on disk that we wanted to
***   write to. This can also happen when, by chance, the move's source and
***   target ranges overlap.
*** - In response, we can break the extent down into sub-sections and
***   attempt to wipe each subsection (eventually down to a granularity
***   of one cluster). We also inspect allocated/free sectors to look ahead
***   and avoid making move calls that we know will fail.
*** - If a cluster was allocated by some other Windows process before we could
***   explicitly wipe it, it is assumed to be wiped. Even if Windows writes a
***   small amount of explicit data to a cluster, it seems to write zero-fill
***   out to the end of the cluster to round it out.
***
***
***   TO DO
***   - Test working correctly if per-user disk quotas are in place
***
"""


# standard library
import sys
import os
import struct
import logging
from operator import itemgetter
from random import randint
from collections import namedtuple

# third-party
# pylint: disable=no-name-in-module
from win32api import (GetVolumeInformation, GetDiskFreeSpace,
                      GetVersionEx, Sleep)
from win32file import (CreateFile, CreateFileW,
                       CloseHandle, GetDriveType,
                       GetFileSize, GetFileAttributesW,
                       SetFileAttributesW,
                       DeviceIoControl, SetFilePointer,
                       WriteFile,
                       LockFile, DeleteFile,
                       SetEndOfFile, FlushFileBuffers)
from winioctlcon import (FSCTL_GET_RETRIEVAL_POINTERS,
                         FSCTL_GET_VOLUME_BITMAP,
                         FSCTL_GET_NTFS_VOLUME_DATA,
                         FSCTL_MOVE_FILE,
                         FSCTL_SET_COMPRESSION,
                         FSCTL_SET_SPARSE,
                         FSCTL_SET_ZERO_DATA)
from win32file import (GENERIC_READ, GENERIC_WRITE, FILE_BEGIN,
                       FILE_SHARE_DELETE,
                       FILE_SHARE_READ, FILE_SHARE_WRITE,
                       OPEN_EXISTING, CREATE_ALWAYS, FILE_FLAG_BACKUP_SEMANTICS,
                       DRIVE_REMOTE, DRIVE_CDROM, DRIVE_UNKNOWN)
from win32con import (FILE_ATTRIBUTE_ENCRYPTED,
                      FILE_ATTRIBUTE_COMPRESSED,
                      FILE_ATTRIBUTE_SPARSE_FILE,
                      FILE_ATTRIBUTE_HIDDEN,
                      FILE_ATTRIBUTE_READONLY,
                      FILE_FLAG_RANDOM_ACCESS,
                      FILE_FLAG_NO_BUFFERING,
                      FILE_FLAG_WRITE_THROUGH,
                      COMPRESSION_FORMAT_DEFAULT)

# local import
from bleachbit.FileUtilities import extended_path, extended_path_undo

# Constants.
VER_SUITE_PERSONAL = 0x200   # doesn't seem to be present in win32con.
SIMULATE_CONCURRENCY = False     # remove this test function when QA complete
# drive_letter_safety = "E"       # protection to only use removable drives
# don't use C: or D:, but E: and beyond OK.
TMP_FILE_NAME = "bbtemp.dat"
SPIKE_FILE_NAME = "bbspike"     # cluster number will be appended
WRITE_BUF_SIZE = 512 * 1024     # 512 kilobytes
ZERO_FILL_BUFFER = bytearray(WRITE_BUF_SIZE)

# Set up logging
logger = logging.getLogger(__name__)


def unpack_element(fmt, structure):
    """Unpacks the next element in a structure, using format requested.

    Args:
        fmt: Format string for struct.unpack
        structure: Structure to unpack

    Returns:
        tuple: Element and remaining content of the structure
    """
    chunk_size = struct.calcsize(fmt)
    element = struct.unpack(fmt, structure[:chunk_size])
    if element and len(element) > 0:
        element = element[0]    # convert from tuple to single element
    structure = structure[chunk_size:]
    return element, structure

# Convert VCN/LCN tuples into cluster start/end tuples.


def logical_ranges_to_extents(ranges, bridge_compressed=False):
    """Convert a list of VCN/LCN tuples into a list of cluster start/end tuples.

    Args:
        ranges: List of VCN/LCN tuples from GET_RETRIEVAL_POINTERS
        bridge_compressed: If True, combines nearly contiguous extents in compressed files
                           even if there are unrelated clusters between them

    Yields:
        Tuples of (start_cluster, end_cluster) representing extents on disk
    """
    if not bridge_compressed:
        vcn_count = 0
        for vcn, lcn in ranges:
            # If we encounter an LCN of -1, we have reached a
            # "space-saved" part of a compressed file. These clusters
            # don't map to clusters on disk, just advance beyond them.
            if lcn < 0:
                vcn_count = vcn
                continue

            # Figure out length for this cluster range.
            # Keep track of VCN inside this file.
            this_vcn_span = vcn - vcn_count
            vcn_count = vcn
            assert this_vcn_span >= 0

            yield (lcn, lcn + this_vcn_span - 1)

    else:
        vcn_count = 0
        last_record = len(ranges)
        index = 0
        while index < last_record:
            vcn, lcn = ranges[index]

            # If we encounter an LCN of -1, we have reached a
            # "space-saved" part of a compressed file. These clusters
            # don't map to clusters on disk, just advance beyond them.
            if lcn < 0:
                vcn_count = vcn
                index += 1
                continue

            # Figure out if we have a block of clusters that can
            # be merged together. The pattern is regular disk
            # clusters interspersed with -1 space-saver sections
            # that are arranged with gaps of 16 clusters or less.
            merge_index = index
            while (lcn >= 0 and
                   merge_index + 2 < last_record and
                   ranges[merge_index + 1][1] < 0 and
                   ranges[merge_index + 2][1] >= 0 and
                   ranges[merge_index + 2][1] - ranges[merge_index][1] <= 16 and
                   ranges[merge_index + 2][1] - ranges[merge_index][1] > 0):
                merge_index += 2

            # Figure out length for this cluster range.
            # Keep track of VCN inside this file.
            if merge_index == index:
                index += 1
                this_vcn_span = vcn - vcn_count
                vcn_count = vcn
                assert this_vcn_span >= 0
                yield (lcn, lcn + this_vcn_span - 1)
            else:
                index = merge_index + 1
                last_vcn_span = (ranges[merge_index][0] -
                                 ranges[merge_index - 1][0])
                vcn = ranges[merge_index][0]
                vcn_count = vcn
                assert last_vcn_span >= 0
                yield (lcn, ranges[merge_index][1] + last_vcn_span - 1)


# Determine clusters that are in extents list A but not in B.
# Generator function, will return results one tuple at a time.
def extents_a_minus_b(a, b):
    """Calculate clusters that exist in extents list A but not in B.

    Args:
        a: List of tuples (start_cluster, end_cluster) representing extents
        b: List of tuples (start_cluster, end_cluster) to exclude from a

    Yields:
        Tuples of (start_cluster, end_cluster) for extents in a but not in b
    """
    # Sort the lists of start/end points.
    a_sorted = sorted(a, key=itemgetter(0))
    b_sorted = sorted(b, key=itemgetter(0))
    b_is_empty = not b

    for a_begin, a_end in a_sorted:
        # If B is an empty list, each item of A will be unchanged.
        if b_is_empty:
            yield (a_begin, a_end)

        for b_begin, b_end in b_sorted:
            if b_begin > a_end:
                # Already gone beyond current A range and no matches.
                # Return this range of A unbroken.
                yield (a_begin, a_end)
                break
            elif b_end < a_begin:
                # Too early in list, keep searching.
                continue
            elif b_begin <= a_begin:
                if b_end >= a_end:
                    # This range of A is completely covered by B.
                    # Do nothing and pass on to next range of A.
                    break
                else:
                    # This range of A is partially covered by B.
                    # Remove the covered range from A and loop
                    a_begin = b_end + 1
            else:
                # This range of A is partially covered by B.
                # Return the first part of A not covered.
                # Either process remainder of A range or move to next A.
                yield (a_begin, b_begin - 1)
                if b_end >= a_end:
                    break
                else:
                    a_begin = b_end + 1


def choose_if_bridged(volume_handle, total_clusters,
                      orig_extents, bridged_extents):
    """Decide if it will be more efficient to bridge the extents and wipe
    some additional clusters that weren't strictly part of the file.

    Args:
        volume_handle: Handle to the volume
        total_clusters: Total number of clusters on the volume
        orig_extents: Original extents of the file
        bridged_extents: Bridged extents of the file

    Returns:
        List of tuples (start_cluster, end_cluster) for extents to wipe
    """
    logger.debug('bridged extents: %s', bridged_extents)
    allocated_extents = []
    volume_bitmap, _bitmap_size = get_volume_bitmap(volume_handle,
                                                    total_clusters)
    _count_ofree, count_oallocated = check_extents(
        orig_extents, volume_bitmap)
    _count_bfree, count_ballocated = check_extents(
        bridged_extents,
        volume_bitmap,
        allocated_extents)
    bridged_extents = [x for x in extents_a_minus_b(bridged_extents,
                                                    allocated_extents)]

    extra_allocated_clusters = count_ballocated - count_oallocated
    saving_in_extents = len(orig_extents) - len(bridged_extents)
    logger.debug("Bridged extents would require us to work around %d allocated clusters.",
                 extra_allocated_clusters)
    logger.debug("It would reduce extent count from %d to %d.",
                 len(orig_extents), len(bridged_extents))

    # Use a penalty of 10 extents for each extra allocated cluster.
    # Why 10? Assuming our next granularity above 1 cluster is a 10 cluster
    # extent, a single allocated cluster would cause us to perform 8
    # additional write/move cycles due to splitting that extent into single
    # clusters.
    # If we had a notion of distribution of extra allocated clusters,
    # we could make this calc more exact. But it's just a rule of thumb.
    tradeoff = saving_in_extents - extra_allocated_clusters * 10
    if tradeoff > 0:
        logger.debug("Quickest method should be bridged extents")
        return bridged_extents

    logger.debug("Quickest method should be original extents")
    return orig_extents


def split_extent(lcn_start, lcn_end):
    """Break an extent into smaller portions.

    Args:
        lcn_start: Start of the extent
        lcn_end: End of the extent

    Yields:
        Tuples of (start_cluster, end_cluster) for extents to wipe
    """
    split_factor = 10

    exponent = 0
    count = lcn_end - lcn_start + 1
    while count > split_factor**(exponent + 1.3):
        exponent += 1
    extent_size = split_factor**exponent
    for x in range(lcn_start, lcn_end + 1, extent_size):
        yield (x, min(x + extent_size - 1, lcn_end))


def check_extents(extents, volume_bitmap, allocated_extents=None):
    """Check extents to see if they are marked as free.

    Args:
        extents: List of tuples (start_cluster, end_cluster) representing extents
        volume_bitmap: Bitmap of clusters on the volume
        allocated_extents: List to store allocated extents (optional)

    Returns:
        Tuple of (count_free, count_allocated)
    """
    count_free, count_allocated = (0, 0)
    for lcn_start, lcn_end in extents:
        for cluster in range(lcn_start, lcn_end + 1):
            if check_mapped_bit(volume_bitmap, cluster):
                count_allocated += 1
                if allocated_extents is not None:
                    # Modified by Marvin [12/05/2020] The extents should have (start, end) format
                    allocated_extents.append((cluster, cluster))
            else:
                count_free += 1

    logger.debug("Extents checked: clusters free %d; allocated %d",
                 count_free, count_allocated)
    return (count_free, count_allocated)


# Simulate concurrency for testing by occasionally allocating clusters during checking.
def check_extents_concurrency(extents, volume_bitmap,
                              tmp_file_path, volume_handle,
                              total_clusters,
                              allocated_extents=None):
    """Check extents to see if they are marked as free, with concurrency simulation.

    Like check_extents(), but simulates a concurrent process.

    Args:
        extents: List of tuples (start_cluster, end_cluster) representing extents
        volume_bitmap: Bitmap of clusters on the volume
        tmp_file_path: Path to a temporary file
        volume_handle: Handle to the volume
        total_clusters: Total number of clusters on the volume
        allocated_extents: List to store allocated extents (optional)

    Returns:
        Tuple of (count_free, count_allocated)
    """
    odds_to_allocate = 1200    # 1 in 1200

    count_free, count_allocated = (0, 0)
    for lcn_start, lcn_end in extents:
        for cluster in range(lcn_start, lcn_end + 1):
            # Every once in a while, occupy a particular cluster on disk.
            if randint(1, odds_to_allocate) == odds_to_allocate:
                spike_cluster(volume_handle, cluster, tmp_file_path)
                if bool(randint(0, 1)):
                    # Simulate allocated before the check, by refetching
                    # the volume bitmap.
                    logger.debug("Simulate known allocated")
                    volume_bitmap, _ = get_volume_bitmap(
                        volume_handle, total_clusters)
                else:
                    # Simulate allocated after the check.
                    logger.debug("Simulate unknown allocated")

            if check_mapped_bit(volume_bitmap, cluster):
                count_allocated += 1
                if allocated_extents is not None:
                    allocated_extents.append(cluster)
            else:
                count_free += 1

    logger.debug("Extents checked: clusters free %d; allocated %d",
                 count_free, count_allocated)
    return (count_free, count_allocated)


def spike_cluster(volume_handle, cluster, tmp_file_path):
    """Allocate a specific cluster on disk by creating a file at that location.

    This is only used for testing concurrency issues.

    Args:
        volume_handle: Handle to the volume
        cluster: The cluster number to allocate
        tmp_file_path: Path used as a reference for creating the spike file

    This function simulates another process grabbing a cluster while our
    algorithm is working, which helps test concurrency handling.
    """
    spike_file_path = os.path.dirname(tmp_file_path)
    if spike_file_path[-1] != os.sep:
        spike_file_path += os.sep
    spike_file_path += SPIKE_FILE_NAME + str(cluster)
    file_handle = CreateFile(spike_file_path,
                             GENERIC_READ | GENERIC_WRITE,
                             FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
                             None, CREATE_ALWAYS, 0, None)
    # 2000 bytes is enough to direct the file to its own cluster and not
    # land entirely in the MFT.
    write_zero_fill(file_handle, 2000)
    move_file(volume_handle, file_handle, 0, cluster, 1)
    CloseHandle(file_handle)
    logger.debug("Spiked cluster %d with %s", cluster, spike_file_path)


def check_mapped_bit(volume_bitmap, lcn):
    """Check if an LCN is allocated (True) or free (False).

    The LCN determines at what index into the bytes/bits structure we
    should look.

    Args:
        volume_bitmap: Bitmap of clusters on the volume
        lcn: Logical Cluster Number to check

    Returns:
        Boolean indicating whether the cluster is allocated
    """
    assert isinstance(lcn, int)
    mapped_bit = volume_bitmap[lcn // 8]
    bit_location = lcn % 8    # zero-based
    if bit_location > 0:
        mapped_bit = mapped_bit >> bit_location
    mapped_bit = mapped_bit & 1
    return mapped_bit > 0


def check_os():
    """Check if the current operating system is Windows NT or later.

    Raises:
        RuntimeError: If not running on Windows NT or later
    """
    if os.name.lower() != "nt":
        raise RuntimeError("This function requires Windows NT or later")


def determine_win_version():
    """Determine which version of Windows we are running.

    Not currently used, except to control encryption test case
    depending on whether it's Windows Home Edition or something higher end.

    Returns:
        Tuple of (version, is_home)
    """
    ver_info = GetVersionEx(1)
    is_home = bool(ver_info[7] & VER_SUITE_PERSONAL)
    if ver_info[:2] == (6, 0):
        return "Vista", is_home
    elif ver_info[0] >= 6:
        return "Later than Vista", is_home
    else:
        return "Something else", is_home


def open_file(file_name, mode=GENERIC_READ):
    """Open the file to get a Windows file handle, ensuring it exists.

    Uses CreateFileW for Unicode support.

    Args:
        file_name: Path to the file to open
        mode: Access mode (default: GENERIC_READ)

    Returns:
        Windows file handle
    """
    file_handle = CreateFileW(file_name, mode, 0, None,
                              OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, None)
    return file_handle


def close_file(file_handle):
    """Close the file handle."""
    CloseHandle(file_handle)


def get_file_basic_info(file_name, file_handle):
    """Get some basic information about a file.

    Args:
        file_name: Path to the file
        file_handle: Windows file handle

    Returns:
        Tuple of (file_attributes, file_size)
    """
    file_attributes = GetFileAttributesW(file_name)
    file_size = GetFileSize(file_handle)
    is_compressed = bool(file_attributes & FILE_ATTRIBUTE_COMPRESSED)
    is_encrypted = bool(file_attributes & FILE_ATTRIBUTE_ENCRYPTED)
    is_sparse = bool(file_attributes & FILE_ATTRIBUTE_SPARSE_FILE)
    is_special = is_compressed | is_encrypted | is_sparse
    if is_special:
        logger.debug('%s: %s %s %s', file_name,
                     'compressed' if is_compressed else '',
                     'encrypted' if is_encrypted else '',
                     'sparse' if is_sparse else '')
    return file_size, is_special


def truncate_file(file_handle):
    """Truncate a file.

    Do this when to release its clusters."""
    SetFilePointer(file_handle, 0, FILE_BEGIN)
    SetEndOfFile(file_handle)
    FlushFileBuffers(file_handle)


def volume_from_file(file_name):
    r"""Given a Windows file path, determine the volume that contains it.

    Append the separator \ to it (more useful for subsequent calls).

    Args:
        file_name: Path to the file

    Returns:
        Volume path
    """
    # strip \\?\
    split_path = os.path.splitdrive(extended_path_undo(file_name))
    volume = split_path[0]
    if volume and volume[-1] != os.sep:
        volume += os.sep
    return volume


class UnsupportedFileSystemError(Exception):
    """An exception for an unsupported file system"""

# Given a volume, get the relevant volume information.
# We are interested in:
# First call: Drive Name; Max Path; File System.
# Second call: Sectors per Cluster; Bytes per Sector; Total # of Clusters.
# Third call: Drive Type.


def get_volume_information(volume):
    """Get volume information.

    Args:
        volume: Volume path

    Returns:
        Volume information
    """
    # If it's a UNC path, raise an error.
    if not volume:
        raise UnsupportedFileSystemError(
            "Only files with a Local File System path can be wiped.")

    result1 = GetVolumeInformation(volume)
    result2 = GetDiskFreeSpace(volume)
    result3 = GetDriveType(volume)

    for drive_enum, error_reason in [
            (DRIVE_REMOTE, "a network drive"),
            (DRIVE_CDROM, "a CD-ROM"),
            (DRIVE_UNKNOWN, "an unknown drive type")]:
        if result3 == drive_enum:
            raise UnsupportedFileSystemError(
                f"This file is on {error_reason} and can't be wiped.")

    # Only NTFS and FAT variations are supported.
    # UDF (file system for CD-RW etc) is not supported.
    if result1[4].upper() == "UDF":
        raise UnsupportedFileSystemError(
            "This file system (UDF) is not supported.")

    volume_info = namedtuple('VolumeInfo', [
        'drive_name', 'max_path', 'file_system',
        'sectors_per_cluster', 'bytes_per_sector', 'total_clusters'])

    return volume_info(result1[0], result1[2], result1[4],
                       result2[0], result2[1], result2[3])


def obtain_readwrite(volume):
    """Get read/write access to a volume.

    Args:
        volume: Volume path

    Returns:
        Volume handle
    """
    # Optional protection that we are running on removable media only.
    assert volume
    # if drive_letter_safety:
    #    drive_containing_file = volume[0].upper()
    #    assert drive_containing_file >= drive_letter_safety.upper()

    volume = '\\\\.\\' + volume
    if volume[-1] == os.sep:
        volume = volume.rstrip(os.sep)

    # We need the FILE_SHARE flags so that this open call can succeed
    # despite something on the volume being in use by another process.
    volume_handle = CreateFile(volume, GENERIC_READ | GENERIC_WRITE,
                               FILE_SHARE_READ | FILE_SHARE_WRITE,
                               None, OPEN_EXISTING,
                               FILE_FLAG_RANDOM_ACCESS |
                               FILE_FLAG_NO_BUFFERING |
                               FILE_FLAG_WRITE_THROUGH,
                               None)
    # logger.debug("Opened volume %s", volume)

    return volume_handle


def get_extents(file_handle, translate_to_extents=True, filename="<unknown>"):
    """Retrieve a list of pointers to the file location on disk.

    If translate_to_extents is False, return the Windows VCN/LCN format.
    If True, do an extra conversion to get a list of extents on disk.

    Args:
        file_handle: Windows file handle
        translate_to_extents: Whether to translate to extents
        filename: Name of the file

    Returns:
        List of extents
    """
    # Assemble input structure and query Windows for retrieval pointers.
    # The input structure is the number 0 as a signed 64 bit integer.
    input_struct = struct.pack('q', 0)
    # 4K, 32K, 256K, 2M step ups in buffer size, until call succeeds.
    # Compressed/encrypted/sparse files tend to have more chopped up extents.
    buf_retry_sizes = [4 * 1024, 32 * 1024, 256 * 1024, 2 * 1024**2]

    rp_struct = None
    for retrieval_pointers_buf_size in buf_retry_sizes:
        try:
            rp_struct = DeviceIoControl(file_handle,
                                        FSCTL_GET_RETRIEVAL_POINTERS,
                                        input_struct,
                                        retrieval_pointers_buf_size)
        except:
            err_info = sys.exc_info()[1]
            err_code = err_info.winerror
            if err_code == 38:     # when file size is 0.
                # (38, 'DeviceIoControl', 'Reached the end of the file.')
                return []
            elif err_code in [122, 234]:  # when buffer not large enough.
                # (122, 'DeviceIoControl',
                # 'The data area passed to a system call is too small.')
                # (234, 'DeviceIoControl', 'More data is available.')
                pass
            else:
                logger.error("Unhandled error code %d in get_extents for file '%s': %s",
                             err_code, filename, str(err_info))
                raise
        else:
            # Call succeeded, break out from for loop.
            break

    if rp_struct is None:
        raise Exception(
            f"Failed to get retrieval pointers for file '{filename}'")

    # At this point we have a FSCTL_GET_RETRIEVAL_POINTERS (rp) structure.
    # Process content of the first part of structure.
    # Separate the retrieval pointers list up front, so we are not making
    # too many string copies of it.
    chunk_size = struct.calcsize('IIq')
    rp_list = rp_struct[chunk_size:]
    rp_struct = rp_struct[:chunk_size]
    record_count, rp_struct = unpack_element('I', rp_struct)    # 4 bytes
    _, rp_struct = unpack_element('I', rp_struct)               # 4 bytes
    starting_vcn, rp_struct = unpack_element('q', rp_struct)    # 8 bytes
    # 4 empty bytes were consumed above.
    # This is for reasons of 64-bit alignment inside structure.

    # If we make the GET_RETRIEVAL_POINTERS request with 0,
    # this should always come back 0.
    assert starting_vcn == 0

    # Populate the extents array with the ranges from rp structure.
    ranges = []
    c = record_count
    i = 0
    chunk_size = struct.calcsize('q')
    buf_size = len(rp_list)
    while c > 0 and i < buf_size:
        next_vcn = struct.unpack_from('q', rp_list, offset=i)
        lcn = struct.unpack_from('q', rp_list, offset=i + chunk_size)
        ranges.append((next_vcn[0], lcn[0]))
        i += chunk_size * 2
        c -= 1

    if translate_to_extents:
        return list(logical_ranges_to_extents(ranges))
    return ranges


def file_make_compressed(file_handle):
    """Make this file compressed on disk.

    Used only for test suite.

    Args:
        file_handle: Windows file handle
    """
    # Assemble input structure.
    # Just tell Windows to use standard compression.
    input_struct = struct.pack('H', COMPRESSION_FORMAT_DEFAULT)
    buf_size = struct.calcsize('H')

    _ = DeviceIoControl(file_handle, FSCTL_SET_COMPRESSION,
                        input_struct, buf_size)


def file_make_sparse(file_handle):
    """Make this file sparse on disk.

    Used only for test suite.

    Args:
        file_handle: Windows file handle
    """
    _ = DeviceIoControl(file_handle, FSCTL_SET_SPARSE, None, None)


def file_add_sparse_region(file_handle, byte_start, byte_end):
    """Add a zero region to a sparse file.

    Used only for test suite.
    """
    # Assemble input structure.
    input_struct = struct.pack('qq', byte_start, byte_end)
    buf_size = struct.calcsize('qq')

    _ = DeviceIoControl(file_handle, FSCTL_SET_ZERO_DATA,
                        input_struct, buf_size)


def get_volume_bitmap(volume_handle, total_clusters):
    """Retrieve a bitmap of whether clusters on disk are free/allocated.

    Args:
        volume_handle: Windows volume handle
        total_clusters: Total number of clusters on the volume

    Returns:
        Tuple of volume bitmap and bitmap size
    """
    # Assemble input structure and query Windows for volume bitmap.
    # The input structure is the number 0 as a signed 64 bit integer.
    input_struct = struct.pack('q', 0)

    # Figure out the buffer size. Add small fudge factor to ensure success.
    buf_size = int(total_clusters / 8) + 16 + 64

    vb_struct = DeviceIoControl(volume_handle, FSCTL_GET_VOLUME_BITMAP,
                                input_struct, buf_size)

    # At this point we have a FSCTL_GET_VOLUME_BITMAP (vb) structure.
    # Process content of the first part of structure.
    # Separate the volume bitmap up front, so we are not making too
    # many string copies of it.
    chunk_size = struct.calcsize('2q')
    volume_bitmap = vb_struct[chunk_size:]
    vb_struct = vb_struct[:chunk_size]
    starting_lcn, vb_struct = unpack_element('q', vb_struct)    # 8 bytes
    bitmap_size, vb_struct = unpack_element('q', vb_struct)     # 8 bytes

    # If we make the GET_VOLUME_BITMAP request with 0,
    # this should always come back 0.
    assert starting_lcn == 0

    # The remaining part of the structure is the actual bitmap.
    return volume_bitmap, bitmap_size


def get_ntfs_volume_data(volume_handle):
    """Retrieve info about an NTFS volume.

    We are mainly interested in the locations of the Master File Table.
    This call is currently not necessary, but has been left in to address any
    future need.

    Args:
        volume_handle: Windows volume handle

    Returns:
        Tuple of Master File Table start and end
    """
    # 512 bytes will be comfortably enough to store return object.
    vd_struct = DeviceIoControl(volume_handle, FSCTL_GET_NTFS_VOLUME_DATA,
                                None, 512)

    # At this point we have a FSCTL_GET_NTFS_VOLUME_DATA (vd) structure.
    # Pick out the elements from structure that are useful to us.
    _, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    _number_sectors, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    _total_clusters, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    _free_clusters, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    _total_reserved, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    _, vd_struct = unpack_element('4I', vd_struct)    # 4*4 bytes
    _, vd_struct = unpack_element('3q', vd_struct)    # 3*8 bytes
    mft_zone_start, vd_struct = unpack_element('q', vd_struct)     # 8 bytes
    mft_zone_end, vd_struct = unpack_element('q', vd_struct)     # 8 bytes

    # Quick sanity check that we got something reasonable for MFT zone.
    assert mft_zone_start < mft_zone_end and mft_zone_start > 0 and mft_zone_end > 0

    logger.debug("MFT from %d to %d", mft_zone_start, mft_zone_end)
    return mft_zone_start, mft_zone_end


def poll_clusters_freed(volume_handle, total_clusters, orig_extents):
    """Poll to confirm that our clusters were freed.

    Check ten times per second for a duration of seven seconds.
    According to Windows Internals book, it may take several seconds
    until NTFS does a checkpoint and releases the clusters.
    In later versions of Windows, this seems to be instantaneous.

    Args:
        volume_handle: Windows volume handle
        total_clusters: Total number of clusters on the volume
        orig_extents: Original extents of the file

    Returns:
        True if clusters were freed, False otherwise
    """
    polling_duration_seconds = 7
    attempts_per_second = 10

    if not orig_extents:
        return True

    for _ in range(polling_duration_seconds * attempts_per_second):
        volume_bitmap, _bitmap_size = get_volume_bitmap(volume_handle,
                                                        total_clusters)
        count_free, count_allocated = check_extents(
            orig_extents, volume_bitmap)
        # Some inexact measure to determine if our clusters were freed
        # by the OS, knowing that another process may grab some clusters
        # in between our polling attempts.
        if count_free > count_allocated:
            return True
        Sleep(1000 / attempts_per_second)

    return False


def move_file(volume_handle, file_handle, starting_vcn,
              starting_lcn, cluster_count):
    """Move file clusters to a specific location on disk using the Defrag API.

    Args:
        volume_handle: Handle to the volume containing the file
        file_handle: Handle to the file to be moved
        starting_vcn: Starting Virtual Cluster Number within the file
        starting_lcn: Target Logical Cluster Number on disk
        cluster_count: Number of clusters to move

    Raises:
        Exception: If clusters are not free or if the move operation fails
    """
    assert file_handle is not None
    # Assemble input structure for our request.
    # We include a couple of zero ints for 64-bit alignment.
    input_struct = struct.pack('IIqqII', int(file_handle), 0, starting_vcn,
                               starting_lcn, cluster_count, 0)
    _vb_struct = DeviceIoControl(volume_handle, FSCTL_MOVE_FILE,
                                 input_struct, None)


def write_zero_fill(file_handle, write_length_bytes):
    """Write to fill a file with zeroes.

    write_length_bytes: number of bytes to write.

    This function writes a specified number of zero bytes to a file using the
    provided file handle. The process works as follows:

    1. The function uses the current file pointer position (wherever it was set before
       this function was called).
    2. It writes zeros in chunks of up to WRITE_BUF_SIZE bytes (512KB by default)
       to optimize performance.
    3. Each WriteFile operation automatically advances the file pointer by the
       number of bytes written.
    4. The function continues writing until the specified write_length_bytes is reached.
    5. Finally, it flushes the buffers to ensure all data is written to disk.

    This function doesn't explicitly set or move the file pointer - it relies on
    the file pointer being positioned correctly before the function is called and
    the automatic advancement of the pointer during write operations.
    """
    assert len(ZERO_FILL_BUFFER) == WRITE_BUF_SIZE
    assert WRITE_BUF_SIZE > 0
    assert file_handle is not None

    # Loop and perform writes of WRITE_BUF_SIZE bytes or less.
    # Continue until write_length_bytes have been written.
    # There is no need to explicitly move the file pointer while
    # writing. We are writing contiguously.
    while write_length_bytes > 0:
        assert write_length_bytes > 0
        if write_length_bytes >= WRITE_BUF_SIZE:
            write_string = ZERO_FILL_BUFFER
            write_length_bytes -= WRITE_BUF_SIZE
        else:
            write_string = ZERO_FILL_BUFFER[:write_length_bytes]
            write_length_bytes = 0

        # Write buffer to file.
        # The WriteFile operation automatically advances the file pointer
        # by the number of bytes written (bytes_written).
        # logger.debug("Write %d bytes", len(write_string))
        _, bytes_written = WriteFile(file_handle, write_string)
        assert bytes_written == len(write_string)

    # Ensure all data is written to disk, not just cached in memory
    FlushFileBuffers(file_handle)


def wipe_file_direct(file_handle, extents, cluster_size, file_size):
    """Wipe a file by directly writing zeros to its clusters.

    This function overwrites the file's data by writing zeros to each of its
    clusters on disk. The process works as follows:

    1. The file pointer starts at position zero in the file.
    2. For each extent (a contiguous range of clusters), we calculate how many
       bytes to write based on the cluster size and number of clusters.
    3. We write zeros to the file using the current file pointer position.
    4. The file system writes to the original clusters rather than allocating
       new ones because we're writing in-place with an open file handle.
    5. This ensures the original data is completely overwritten on disk.
    6. The file pointer automatically advances after each write operation.

    If the last extent was not originally full, the file size will increase
    to be a multiple of the cluster size.

    If the file had no extents (very small files stored in the MFT),
    we write zeros for the entire file size.
    """
    assert cluster_size > 0

    # Remember that file_size measures full expanded content of the file,
    # which may not always match with size on disk (eg. if file compressed).
    LockFile(file_handle, 0, 0, file_size & 0xFFFF, file_size >> 16)

    if extents:
        # Use size on disk to determine how many clusters of zeros we write.
        for lcn_start, lcn_end in extents:
            # logger.debug("Wiping extent from %d to %d...",
            #              lcn_start, lcn_end)
            # Calculate write length based on the number of clusters in this extent.
            # The file pointer is positioned at the start of the current extent.
            # Each write operation will advance the file pointer automatically.
            write_length = (lcn_end - lcn_start + 1) * cluster_size
    else:
        # Special case - file so small it can be contained within the
        # directory entry in the MFT part of the disk.
        # logger.debug("Wiping tiny file that fits entirely on MFT")
        write_length = file_size
    write_zero_fill(file_handle, write_length)


# Wipe an extent by making calls to the defrag API.
def wipe_extent_by_defrag(volume_handle, lcn_start, lcn_end, cluster_size,
                          total_clusters, tmp_file_path):
    """Wipe disk clusters by creating a zero-filled file and moving it to target location.

    This function uses the Windows Defragmentation API to precisely overwrite specific
    clusters on disk. The process works as follows:

    1. Check if target clusters are free using the volume bitmap
    2. If clusters are allocated or the extent is too large, split into smaller pieces
    3. Create a zero-filled temporary file
    4. Move the temporary file's clusters to the target location on disk

    Args:
        volume_handle: Handle to the volume containing the clusters
        lcn_start: Starting Logical Cluster Number (absolute position on disk)
        lcn_end: Ending Logical Cluster Number (inclusive)
        cluster_size: Size of each cluster in bytes
        total_clusters: Total number of clusters on the volume
        tmp_file_path: Path to use for temporary zero-filled file

    Returns:
        bool: True if wiping succeeded, False otherwise
    """
    assert cluster_size > 0
    logger.debug("Examining extent from %d to %d for wipe...",
                 lcn_start, lcn_end)
    write_length = (lcn_end - lcn_start + 1) * cluster_size

    # Check the state of the volume bitmap for the extent we want to
    # overwrite. If any sectors are allocated, reduce the task
    # into smaller parts.
    # We also reduce to smaller pieces if the extent is larger than
    # 2 megabytes. For no particular reason except to avoid the entire
    # request failing because one cluster became allocated.
    volume_bitmap, _bitmap_size = get_volume_bitmap(volume_handle,
                                                    total_clusters)
    # This option simulates another process that grabs clusters on disk
    # from time to time.
    # It should be moved away after QA is complete.
    if not SIMULATE_CONCURRENCY:
        count_free, count_allocated = check_extents(
            [(lcn_start, lcn_end)], volume_bitmap)
    else:
        count_free, count_allocated = check_extents_concurrency(
            [(lcn_start, lcn_end)], volume_bitmap,
            tmp_file_path, volume_handle, total_clusters)
    if count_allocated > 0 and count_free == 0:
        return False
    if count_allocated > 0 or write_length > WRITE_BUF_SIZE * 4:
        if lcn_start >= lcn_end:
            return False

        for split_s, split_e in split_extent(lcn_start, lcn_end):
            wipe_extent_by_defrag(volume_handle, split_s, split_e,
                                  cluster_size, total_clusters,
                                  tmp_file_path)
        return True
    # Put the zero-fill file in place.
    file_handle = CreateFile(tmp_file_path, GENERIC_READ | GENERIC_WRITE,
                             0, None, CREATE_ALWAYS,
                             FILE_ATTRIBUTE_HIDDEN, None)
    write_zero_fill(file_handle, write_length)
    new_extents = get_extents(file_handle)

    # We know the original extent was contiguous.
    # The new zero-fill file may not be contiguous, so it requires a
    # loop to be sure of reaching the end of the new file's clusters.
    new_vcn = 0
    for new_lcn_start, new_lcn_end in new_extents:
        # logger.debug("Zero-fill wrote from %d to %d",
        #              new_lcn_start, new_lcn_end)
        cluster_count = new_lcn_end - new_lcn_start + 1
        cluster_dest = lcn_start + new_vcn

        if new_lcn_start != cluster_dest:
            logger.debug("Move %d clusters to %d",
                         cluster_count, cluster_dest)
            try:
                move_file(volume_handle, file_handle, new_vcn,
                          cluster_dest, cluster_count)
            except:
                # Move file failed, probably because another process
                # has allocated a cluster on disk.
                # Break into smaller pieces and do what we can.
                logger.debug("!! Move encountered an error !!")
                CloseHandle(file_handle)
                if lcn_start >= lcn_end:
                    return False
                for split_s, split_e in split_extent(lcn_start, lcn_end):
                    wipe_extent_by_defrag(volume_handle, split_s, split_e,
                                          cluster_size, total_clusters,
                                          tmp_file_path)
                return True
        else:
            # If Windows put the zero-fill extent on the exact clusters we
            # intended to place it, no need to attempt a move.
            logging.debug("No need to move extent from %d",
                          new_lcn_start)
        new_vcn += cluster_count

    CloseHandle(file_handle)
    DeleteFile(tmp_file_path)
    return True


def clean_up(file_handle, volume_handle, tmp_file_path):
    """Safely close open handles and delete temporary files.

    Args:
        file_handle: Handle to an open file
        volume_handle: Handle to an open volume
        tmp_file_path: Path to a temporary file that should be deleted

    This function handles exceptions internally to ensure cleanup operations
    continue even if individual operations fail.
    """
    try:
        if file_handle:
            CloseHandle(file_handle)
        if volume_handle:
            CloseHandle(volume_handle)
        if tmp_file_path:
            DeleteFile(tmp_file_path)
    except:
        pass


def file_wipe(file_name):
    """Main function to wipe a file.

    Args:
        file_name: Path to the file to be wiped

    This function handles the entire wiping process, including:
    1. Opening the file and volume
    2. Obtaining file and volume information
    3. Obtaining original extents of the file
    4. Wiping the file
    5. Closing handles and deleting temporary files

    Raises:
        Exception: If any step fails, the function will raise an exception
    """
    # add \\?\ if it does not exist to support Unicode and long paths
    file_name = extended_path(file_name)
    check_os()
    # win_version, _ = determine_win_version()

    volume = volume_from_file(file_name)
    volume_info = get_volume_information(volume)
    cluster_size = (volume_info.sectors_per_cluster *
                    volume_info.bytes_per_sector)

    file_handle = open_file(file_name)
    file_size, is_special = get_file_basic_info(file_name, file_handle)
    orig_extents = get_extents(file_handle, True, file_name)
    if is_special:
        bridged_extents = list(logical_ranges_to_extents(
            get_extents(file_handle, False, file_name), True))
    CloseHandle(file_handle)
    # logger.debug('Original extents: %s', orig_extents)

    volume_handle = obtain_readwrite(volume)
    attrs = GetFileAttributesW(file_name)
    if attrs & FILE_ATTRIBUTE_READONLY:
        # Remove read-only attribute to avoid "access denied" in CreateFileW().
        SetFileAttributesW(file_name, attrs & ~FILE_ATTRIBUTE_READONLY)
    file_handle = open_file(file_name, GENERIC_READ | GENERIC_WRITE)

    if not is_special:
        # Direct overwrite when it's a regular file.
        # logger.info("Attempting direct file wipe.")
        wipe_file_direct(file_handle, orig_extents, cluster_size, file_size)
        new_extents = get_extents(file_handle, True, file_name)
        CloseHandle(file_handle)
        # logger.debug('New extents: %s', new_extents)
        if orig_extents == new_extents:
            clean_up(None, volume_handle, None)
            return
        # Expectation was that extents should be identical and file is wiped.
        # If OS didn't give that to us, continue below and use defrag wipe.
        # Any extent within new_extents has now been wiped by above.
        # It can be subtracted from the orig_extents list, and now we will
        # just clean up anything not yet overwritten.
        orig_extents = extents_a_minus_b(orig_extents, new_extents)
    else:
        # File needs special treatment. We can't just do a basic overwrite.
        # First we will truncate it. Then chase down the freed clusters to
        # wipe them, now that they are no longer part of the file.
        truncate_file(file_handle)
        CloseHandle(file_handle)

    # Poll to confirm that our clusters were freed.
    poll_clusters_freed(volume_handle, volume_info.total_clusters,
                        orig_extents)

    # Chase down all the freed clusters we can, and wipe them.
    # logger.debug("Attempting defrag file wipe.")
    # Put the temp file in the same folder as the target wipe file.
    # Should be able to write this path if user can write the wipe file.
    tmp_file_path = os.path.dirname(file_name) + os.sep + TMP_FILE_NAME
    if is_special:
        orig_extents = choose_if_bridged(volume_handle,
                                         volume_info.total_clusters,
                                         orig_extents, bridged_extents)
    for lcn_start, lcn_end in orig_extents:
        wipe_extent_by_defrag(volume_handle, lcn_start, lcn_end,
                              cluster_size, volume_info.total_clusters,
                              tmp_file_path)

    # Clean up.
    clean_up(None, volume_handle, tmp_file_path)
    return



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
GUI for making chaff
"""

from bleachbit.Language import get_text as _

from gi.repository import Gtk, GLib
import logging
import os


logger = logging.getLogger(__name__)


def make_files_thread(file_count, inspiration, output_folder, delete_when_finished, on_progress):
    if inspiration == 0:
        from bleachbit.Chaff import generate_2600
        generated_file_names = generate_2600(
            file_count, output_folder, on_progress=on_progress)
    elif inspiration == 1:
        from bleachbit.Chaff import generate_emails
        generated_file_names = generate_emails(
            file_count, output_folder, on_progress=on_progress)
    else:
        raise ValueError(f'Invalid inspiration {inspiration}')
    if delete_when_finished:
        on_progress(0, msg=_('Deleting files'))
        for i in range(0, file_count):
            os.unlink(generated_file_names[i])
            on_progress(1.0 * (i + 1) / file_count)
    on_progress(1.0, is_done=True)


class ChaffDialog(Gtk.Dialog):

    """Present the dialog to make chaff"""

    def __init__(self, parent):
        self._make_dialog(parent)

    def _make_dialog(self, parent):
        """Make the main dialog"""
# TRANSLATORS: BleachBit creates digital chaff like that is like the
# physical chaff airplanes use to protect themselves from radar-guided
# missiles. For more explanation, see the online documentation.
        Gtk.Dialog.__init__(self, title=_("Make chaff"), transient_for=parent)
        Gtk.Dialog.set_modal(self, True)
        self.set_border_width(5)
        box = self.get_content_area()

        label = Gtk.Label(
            label=_("Make randomly-generated messages inspired by documents."))
        box.add(label)

        inspiration_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        # TRANSLATORS: Inspiration is a choice of documents from which random
        # text will be generated.
        inspiration_box.add(Gtk.Label(label=_("Inspiration")))
        self.inspiration_combo = Gtk.ComboBoxText()
        self.inspiration_combo_options = (
            _('2600 Magazine'), _("Hillary Clinton's emails"))
        for combo_option in self.inspiration_combo_options:
            self.inspiration_combo.append_text(combo_option)
        self.inspiration_combo.set_active(0)  # Set default
        inspiration_box.add(self.inspiration_combo)
        box.add(inspiration_box)

        spin_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        spin_box.add(Gtk.Label(label=_("Number of files")))
        adjustment = Gtk.Adjustment(
            value=100, lower=1, upper=99999, step_increment=1, page_increment=1000, page_size=0)
        self.file_count = Gtk.SpinButton(adjustment=adjustment)
        spin_box.add(self.file_count)
        box.add(spin_box)

        folder_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        folder_box.add(Gtk.Label(label=_("Select destination folder")))
        # The file chooser button displays a stock GTK icon. When some parts of GTK are not
        # set up correctly on Windows, then the application may crash here with the error
        # message "No GSettings schemas".
        # https://github.com/bleachbit/bleachbit/issues/1780
        self.choose_folder_button = Gtk.FileChooserButton()
        self.choose_folder_button.set_action(
            Gtk.FileChooserAction.SELECT_FOLDER)
        import tempfile
        self.choose_folder_button.set_filename(tempfile.gettempdir())
        folder_box.add(self.choose_folder_button)
        box.add(folder_box)

        delete_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        delete_box.add(Gtk.Label(label=_("When finished")))
        self.when_finished_combo = Gtk.ComboBoxText()
        self.combo_options = (
            _('Delete without shredding'), _('Do not delete'))
        for combo_option in self.combo_options:
            self.when_finished_combo.append_text(combo_option)
        self.when_finished_combo.set_active(0)  # Set default
        delete_box.add(self.when_finished_combo)
        box.add(delete_box)

        self.progressbar = Gtk.ProgressBar()
        box.add(self.progressbar)
        self.progressbar.hide()

        self.make_button = Gtk.Button(label=_("Make files"))
        self.make_button.connect('clicked', self.on_make_files)
        box.add(self.make_button)

    def download_models_gui(self):
        """Download models and return whether successful as boolean"""
        def on_download_error(msg, msg2):
            dialog = Gtk.MessageDialog(self, 0, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.CANCEL, msg)
            dialog.format_secondary_text(msg2)
            dialog.run()
            dialog.destroy()
        import bleachbit.Chaff
        return bleachbit.Chaff.download_models(on_error=on_download_error)

    def download_models_dialog(self):
        """Download models"""
        dialog = Gtk.MessageDialog(parent=self, flags=0, message_type=Gtk.MessageType.QUESTION,
                                   buttons=Gtk.ButtonsType.OK_CANCEL, text=_("Download data needed for chaff generator?"))
        response = dialog.run()
        ret = None
        if response == Gtk.ResponseType.OK:
            # User wants to download
            ret = self.download_models_gui()  # True if successful
        elif response == Gtk.ResponseType.CANCEL:
            ret = False
        dialog.destroy()
        return ret

    def on_make_files(self, widget):
        """Callback for make files button"""
        file_count = self.file_count.get_value_as_int()
        output_dir = self.choose_folder_button.get_filename()
        delete_when_finished = self.when_finished_combo.get_active() == 0
        inspiration = self.inspiration_combo.get_active()
        if not output_dir:
            dialog = Gtk.MessageDialog(self, 0, Gtk.MessageType.ERROR,
                                       Gtk.ButtonsType.CANCEL, _("Select destination folder"))
            dialog.run()
            dialog.destroy()
            return

        from bleachbit.Chaff import have_models
        if not have_models():
            if not self.download_models_dialog():
                return

        def _on_progress(fraction, msg, is_done):
            """Update progress bar from GLib main loop"""
            if msg:
                self.progressbar.set_text(msg)
            self.progressbar.set_fraction(fraction)
            if is_done:
                self.progressbar.hide()
                self.make_button.set_sensitive(True)

        def on_progress(fraction, msg=None, is_done=False):
            """Callback for progress bar"""
            # Use idle_add() because threads cannot make GDK calls.
            GLib.idle_add(_on_progress, fraction, msg, is_done)

        msg = _('Generating files')
        logger.info(msg)
        self.progressbar.show()
        self.progressbar.set_text(msg)
        self.progressbar.set_show_text(True)
        self.progressbar.set_fraction(0.0)
        self.make_button.set_sensitive(False)
        import threading
        args = (file_count, inspiration, output_dir,
                delete_when_finished, on_progress)
        self.thread = threading.Thread(target=make_files_thread, args=args)
        self.thread.start()

    def run(self):
        """Run the dialog"""
        self.show_all()



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import bz2
from datetime import datetime
import email.generator
from email.mime.text import MIMEText
import json
import logging
import os
import random
import tempfile


from bleachbit import options_dir


from . import markovify

logger = logging.getLogger(__name__)

# These were typos in the original emails, not OCR errors:
# abdinh@state.gov
# mhcaleja@state.gov

RECIPIENTS = [
    'abedinh@state.gov',
    'adlerce@state.gov',
    'baerdb@state.gov',
    'baldersonkm@state.gov',
    'balderstonkm@state.gov',
    'bam@mikulski.senate.gov',
    'bealeca@state.gov',
    'benjamin_moncrief@lemieux.senate.gov',
    'blaker2@state.gov',
    'brimmere@state.gov',
    'burnswj@state.gov',
    'butzgych2@state.gov',
    'campbellkm@state.gov',
    'carsonj@state.gov',
    'cholletdh@state.gov',
    'cindy.buhl@mail.house.gov',
    'colemancl@state.gov',
    'crowleypj@state.gov',
    'danieljj@state.gov',
    'david_garten@lautenberg.senate.gov',
    'dewanll@state.gov',
    'feltmanjd@state.gov',
    'fuchsmh@state.gov',
    'goldbergps@state.gov',
    'goldenjr@state.gov',
    'gonzalezjs@state.gov',
    'gordonph@state.gov',
    'hanleymr@state.gov',
    'hdr22@clintonemail.com',
    'hillcr@state.gov',
    'holbrookerc@state.gov',
    'hormatsrd@state.gov',
    'hr15@att.blackberry.net',
    'hr15@mycingular.blackberry.net',
    'hrod17@clintonemail.com',
    'huma@clintonemail.com',
    'hyded@state.gov',
    'info@mailva.evite.com',
    'jilotylc@state.gov',
    'jonespw2@state.gov',
    'kellyc@state.gov',
    'klevorickcb@state.gov',
    'kohhh@state.gov',
    'laszczychj@state.gov',
    'lewjj@state.gov',
    'macmanusje@state.gov',
    'marshallcp@state.gov',
    'mchaleja@state.gov',
    'millscd@state.gov',
    'muscatinel@state.gov',
    'nidestr@state.gov',
    'nulandvj@state.gov',
    'oterom2@state.gov',
    'posnermh@state.gov',
    'reinesp@state.gov',
    'reinespi@state.gov',
    'ricese@state.gov',
    'rodriguezme@state.gov',
    'rooneym@state.gov',
    's_specialassistants@state.gov',
    'schwerindb@state.gov',
    'shannonta@state.gov',
    'shapiroa@state.gov',
    'shermanwr@state.gov',
    'slaughtera@state.gov',
    'steinbergjb@state.gov',
    'sterntd@state.gov',
    'sullivanjj@state.gov',
    'tauschereo@state.gov',
    'tillemannts@state.gov',
    'toivnf@state.gov',
    'tommy_ross@reid.senate.gov',
    'valenzuelaaa@state.gov',
    'valmorolj@state.gov',
    'vermarr@state.gov',
    'verveerms@state.gov',
    'woodardew@state.gov']
DEFAULT_SUBJECT_LENGTH = 64
DEFAULT_NUMBER_OF_SENTENCES_CLINTON = 50
DEFAULT_NUMBER_OF_SENTENCES_2600 = 50
MODEL_BASENAMES = (
    '2600_model.json.bz2',
    'clinton_content_model.json.bz2',
    'clinton_subject_model.json.bz2')
URL_TEMPLATES = (
    'https://sourceforge.net/projects/bleachbit/files/chaff/%s/download',
    'https://download.bleachbit.org/chaff/%s')
DEFAULT_MODELS_DIR = options_dir


def _load_model(model_path):
    _open = open
    if model_path.endswith('.bz2'):
        _open = bz2.open
    with _open(model_path, 'rt', encoding='utf-8') as model_file:
        return markovify.Text.from_dict(json.load(model_file))


def load_subject_model(model_path):
    return _load_model(model_path)


def load_content_model(model_path):
    return _load_model(model_path)


def load_2600_model(model_path):
    return _load_model(model_path)


def _get_random_recipient():
    return random.choice(RECIPIENTS)


def _get_random_datetime(min_year=2011, max_year=2012):
    date = datetime.strptime('{} {}'.format(random.randint(
        1, 365), random.randint(min_year, max_year)), '%j %Y')
    # Saturday, September 15, 2012 2:20 PM
    return date.strftime('%A, %B %d, %Y %I:%M %p')


def _get_random_content(content_model, number_of_sentences=DEFAULT_NUMBER_OF_SENTENCES_CLINTON):
    content = []
    for _i in range(number_of_sentences):
        content.append(content_model.make_sentence())
        content.append(random.choice([' ', ' ', '\n\n']))
    try:
        return MIMEText(''.join(content), _charset='iso-8859-1')
    except UnicodeEncodeError:
        return _get_random_content(content_model, number_of_sentences=number_of_sentences)


def _generate_email(subject_model, content_model, number_of_sentences=DEFAULT_NUMBER_OF_SENTENCES_CLINTON, subject_length=DEFAULT_SUBJECT_LENGTH):
    message = _get_random_content(
        content_model, number_of_sentences=number_of_sentences)

    message['Subject'] = subject_model.make_short_sentence(subject_length)
    message['To'] = _get_random_recipient()
    message['From'] = _get_random_recipient()
    message['Sent'] = _get_random_datetime()

    return message


def download_models(models_dir=DEFAULT_MODELS_DIR,
                    on_error=None):
    """Download models

    Calls on_error(primary_message, secondary_message) in case of error

    Returns success as boolean value
    """
    from bleachbit.Network import download_url_to_fn
    for basename in (MODEL_BASENAMES):
        fn = os.path.join(models_dir, basename)
        if os.path.exists(fn):
            logger.debug('File %s already exists', fn)
            continue
        this_file_success = False
        for url_template in URL_TEMPLATES:
            url = url_template % basename
            if download_url_to_fn(url, fn, on_error=on_error):
                this_file_success = True
                break
        if not this_file_success:
            return False
    return True


def generate_emails(number_of_emails,
                    email_output_dir,
                    models_dir=DEFAULT_MODELS_DIR,
                    number_of_sentences=DEFAULT_NUMBER_OF_SENTENCES_CLINTON,
                    on_progress=None,
                    *kwargs):
    logger.debug('Loading two email models')
    subject_model_path = os.path.join(
        models_dir, 'clinton_subject_model.json.bz2')
    content_model_path = os.path.join(
        models_dir, 'clinton_content_model.json.bz2')
    subject_model = load_subject_model(subject_model_path)
    content_model = load_content_model(content_model_path)
    logger.debug('Generating {:,} emails'.format(number_of_emails))
    generated_file_names = []
    for i in range(1, number_of_emails + 1):
        with tempfile.NamedTemporaryFile(mode='w+', prefix='outlook-', suffix='.eml', dir=email_output_dir, delete=False) as email_output_file:
            email_generator = email.generator.Generator(email_output_file)
            msg = _generate_email(
                subject_model, content_model, number_of_sentences=number_of_sentences)
            email_generator.write(msg.as_string())
            generated_file_names.append(email_output_file.name)
        if on_progress:
            on_progress(1.0 * i / number_of_emails)
    return generated_file_names


def _generate_2600_file(model, number_of_sentences=DEFAULT_NUMBER_OF_SENTENCES_2600):
    content = []
    for _i in range(number_of_sentences):
        content.append(model.make_sentence())
        # The space is repeated to make paragraphs longer.
        content.append(random.choice([' ', ' ', '\n\n']))
    return ''.join(content)


def generate_2600(file_count,
                  output_dir,
                  model_dir=DEFAULT_MODELS_DIR,
                  on_progress=None):
    logger.debug('Loading 2600 model')
    model_path = os.path.join(model_dir, '2600_model.json.bz2')
    model = _load_model(model_path)
    logger.debug(f'Generating {file_count:,} files')
    generated_file_names = []
    for i in range(1, file_count + 1):
        with tempfile.NamedTemporaryFile(mode='w+', encoding='utf-8', prefix='2600-', suffix='.txt', dir=output_dir, delete=False) as output_file:
            txt = _generate_2600_file(model)
            output_file.write(txt)
            generated_file_names.append(output_file.name)
        if on_progress:
            on_progress(1.0 * i / file_count)
    return generated_file_names


def have_models():
    """Check whether the models exist in the default location.

    Used to check whether download is needed."""
    for basename in (MODEL_BASENAMES):
        fn = os.path.join(DEFAULT_MODELS_DIR, basename)
        if not os.path.exists(fn):
            return False
    return True



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Code that is commonly shared throughout BleachBit
"""

import os
import re
import sys

from bleachbit import Log
from configparser import RawConfigParser, NoOptionError  # used in other files

APP_VERSION = "5.0.0"
APP_NAME = "BleachBit"
APP_URL = "https://www.bleachbit.org"
APP_COPYRIGHT = "Copyright (C) 2008-2025 Andrew Ziem"

socket_timeout = 10

if sys.version_info < (3, 8, 0):
    print('BleachBit requires Python version 3.8 or later')
    sys.exit(1)

if hasattr(sys, 'frozen') and sys.frozen == 'windows_exe':
    stdout_encoding = 'utf-8'
else:
    stdout_encoding = sys.stdout.encoding

logger = Log.init_log()

# Setting below value to false disables update notification (useful
# for packages in repositories).
online_update_notification_enabled = True

#
# Paths
#

# Windows
bleachbit_exe_path = None
if hasattr(sys, 'frozen'):
    # running frozen in py2exe
    bleachbit_exe_path = os.path.dirname(sys.executable)
else:
    # __file__ is absolute path to __init__.py
    bleachbit_exe_path = os.path.dirname(os.path.dirname(__file__))

# license
license_filename = None
license_filenames = ('/usr/share/common-licenses/GPL-3',  # Debian, Ubuntu
                     # Microsoft Windows
                     os.path.join(bleachbit_exe_path, 'COPYING'),
                     '/usr/share/doc/bleachbit-' + APP_VERSION + '/COPYING',  # CentOS, Fedora, RHEL
                     '/usr/share/licenses/bleachbit/COPYING',  # Fedora 21+, RHEL 7+
                     '/usr/share/doc/packages/bleachbit/COPYING',  # OpenSUSE 11.1
                     '/usr/pkg/share/doc/bleachbit/COPYING',  # NetBSD 5
                     '/usr/share/licenses/common/GPL3/license.txt')  # Arch Linux
for lf in license_filenames:
    if os.path.exists(lf):
        license_filename = lf
        break

# configuration
portable_mode = False
options_dir = None
if 'posix' == os.name:
    options_dir = os.path.expanduser("~/.config/bleachbit")
elif 'nt' == os.name:
    os.environ.pop('FONTCONFIG_FILE', None)
    if os.path.exists(os.path.join(bleachbit_exe_path, 'bleachbit.ini')):
        # portable mode
        portable_mode = True
        options_dir = bleachbit_exe_path
    else:
        # installed mode
        options_dir = os.path.expandvars(r"${APPDATA}\BleachBit")

try:
    options_dir = os.environ['BLEACHBIT_TEST_OPTIONS_DIR']
except KeyError:
    pass

options_file = os.path.join(options_dir, "bleachbit.ini")

# check whether the application is running from the source tree
if not portable_mode:
    paths = (
        '../cleaners',
        '../Makefile',
        '../COPYING')
    existing = (
        os.path.exists(os.path.join(bleachbit_exe_path, path))
        for path in paths)
    portable_mode = all(existing)

# personal cleaners
personal_cleaners_dir = os.path.join(options_dir, "cleaners")

# system cleaners
# On Windows in portable mode, the bleachbit_exe_path is equal to
# options_dir, so be careful that system_cleaner_dir is not set to
# personal_cleaners_dir.
if os.path.isdir(os.path.join(bleachbit_exe_path, 'cleaners')) and not portable_mode:
    system_cleaners_dir = os.path.join(bleachbit_exe_path, 'cleaners')
elif sys.platform in ('linux', 'darwin'):
    system_cleaners_dir = '/usr/share/bleachbit/cleaners'
elif sys.platform == 'win32':
    system_cleaners_dir = os.path.join(bleachbit_exe_path, 'share\\cleaners\\')
elif sys.platform[:6] == 'netbsd':
    system_cleaners_dir = '/usr/pkg/share/bleachbit/cleaners'
elif sys.platform.startswith('openbsd') or sys.platform.startswith('freebsd'):
    system_cleaners_dir = '/usr/local/share/bleachbit/cleaners'
else:
    system_cleaners_dir = None
    logger.warning(
        'unknown system cleaners directory for platform %s ', sys.platform)

# local cleaners directory for running without installation (Windows or Linux)
local_cleaners_dir = None
if portable_mode:
    local_cleaners_dir = os.path.join(bleachbit_exe_path, 'cleaners')

# windows10 theme
windows10_theme_path = os.path.normpath(
    os.path.join(bleachbit_exe_path, 'themes/windows10'))

# application icon
__icons = (
    '/usr/share/pixmaps/bleachbit.png',  # Linux
    '/usr/pkg/share/pixmaps/bleachbit.png',  # NetBSD
    '/usr/local/share/pixmaps/bleachbit.png',  # FreeBSD and OpenBSD
    os.path.normpath(os.path.join(bleachbit_exe_path,
                                  'share\\bleachbit.png')),  # Windows
    # When running from source (i.e., not installed).
    os.path.normpath(os.path.join(bleachbit_exe_path, 'bleachbit.png')),
)
appicon_path = None
for __icon in __icons:
    if os.path.exists(__icon):
        appicon_path = __icon

# menu
# This path works when running from source (cross platform) or when
# installed on Windows.
app_menu_filename = os.path.join(bleachbit_exe_path, 'data', 'app-menu.ui')
if not os.path.exists(app_menu_filename) and system_cleaners_dir:
    # This path works when installed on Linux.
    app_menu_filename = os.path.abspath(
        os.path.join(system_cleaners_dir, '../app-menu.ui'))
if not os.path.exists(app_menu_filename):
    logger.error('unknown location for app-menu.ui')

# locale directory
if os.path.exists("./locale/"):
    # local locale (personal)
    locale_dir = os.path.abspath("./locale/")
# system-wide installed locale
elif sys.platform in ('linux', 'darwin'):
    locale_dir = "/usr/share/locale/"
elif sys.platform == "win32":
    locale_dir = os.path.join(bleachbit_exe_path, "share\\locale\\")
elif sys.platform[:6] == "netbsd":
    locale_dir = "/usr/pkg/share/locale/"
elif sys.platform.startswith("openbsd") or sys.platform.startswith("freebsd"):
    locale_dir = "/usr/local/share/locale/"


#
# URLs
#
base_url = "https://update.bleachbit.org"
help_contents_url = "%s/help/%s" \
    % (base_url, APP_VERSION)
update_check_url = "%s/update/%s" % (base_url, APP_VERSION)

# set up environment variables
if 'nt' == os.name:
    from bleachbit import Windows
    Windows.setup_environment()

if 'posix' == os.name:
    # XDG base directory specification
    envs = {
        'XDG_DATA_HOME': os.path.expanduser('~/.local/share'),
        'XDG_CONFIG_HOME': os.path.expanduser('~/.config'),
        'XDG_CACHE_HOME': os.path.expanduser('~/.cache')
    }
    for varname, value in envs.items():
        if not os.getenv(varname):
            os.environ[varname] = value

# should be re.IGNORECASE on macOS
fs_scan_re_flags = 0 if os.name == 'posix' else re.IGNORECASE

if 'win32' == sys.platform:
    import win32process

    for process in win32process.EnumProcessModules(-1):
        name = win32process.GetModuleFileNameEx(-1, process)
        if re.search(r'python\d+.dll$', name, re.IGNORECASE):
            bindir = os.path.dirname(name)
            os.environ['GDK_PIXBUF_MODULE_FILE'] = os.path.join(
                bindir, 'lib', 'gdk-pixbuf-2.0', '2.10.0', 'loaders.cache')



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import gettext
import os
import logging
import sys

logger = logging.getLogger(__name__)


native_locale_names = \
    {'aa': 'Afaraf',
     'ab': 'аҧсуа бызшәа',
     'ace': 'بهسا اچيه',
     'ach': 'Acoli',
     'ae': 'avesta',
     'af': 'Afrikaans',
     'ak': 'Akan',
     'am': 'አማርኛ',
     'an': 'aragonés',
     'ang': 'Old English',
     'anp': 'Angika',
     'ar': 'العربية',
     'as': 'অসমীয়া',
     'ast': 'Asturianu',
     'av': 'авар мацӀ',
     'ay': 'aymar aru',
     'az': 'azərbaycan dili',
     'ba': 'башҡорт теле',
     'bal': 'Baluchi',
     'be': 'Беларуская мова',
     'bg': 'български език',
     'bh': 'भोजपुरी',
     'bi': 'Bislama',
     'bm': 'bamanankan',
     'bn': 'বাংলা',
     'bo': 'བོད་ཡིག',
     'br': 'brezhoneg',
     'brx': 'Bodo (India)',
     'bs': 'босански',
     'byn': 'Bilin',
     'ca': 'català',
     'ce': 'нохчийн мотт',
     'cgg': 'Chiga',
     'ch': 'Chamoru',
     'ckb': 'Central Kurdish',
     'co': 'corsu',
     'cr': 'ᓀᐦᐃᔭᐍᐏᐣ',
     'crh': 'Crimean Tatar',
     'cs': 'česky',
     'csb': 'Cashubian',
     'cu': 'ѩзыкъ словѣньскъ',
     'cv': 'чӑваш чӗлхи',
     'cy': 'Cymraeg',
     'da': 'dansk',
     'de': 'Deutsch',
     'doi': 'डोगरी; ڈوگرى',
     'dv': 'ދިވެހި',
     'dz': 'རྫོང་ཁ',
     'ee': 'Eʋegbe',
     'el': 'Ελληνικά',  # modern Greek
     'en': 'English',
     'en_AU': 'Australian English',
     'en_CA': 'Canadian English',
     'en_GB': 'British English',
     'en_US': 'United States English',
     'eo': 'Esperanto',
     'es': 'Español',
     'es_419': 'Latin American Spanish',
     'et': 'eesti',
     'eu': 'euskara',
     'fa': 'فارسی',
     'ff': 'Fulfulde',
     'fi': 'suomen kieli',
     'fil': 'Wikang Filipino',
     'fin': 'suomen kieli',
     'fj': 'vosa Vakaviti',
     'fo': 'føroyskt',
     'fr': 'Français',
     'frp': 'Arpitan',
     'fur': 'Frilian',
     'fy': 'Frysk',
     'ga': 'Gaeilge',
     'gd': 'Gàidhlig',
     'gez': 'Geez',
     'gl': 'galego',
     'gn': 'Avañeẽ',
     'grc': 'Ἑλληνική',  # ancient Greek
     'gu': 'Gujarati',
     'gv': 'Gaelg',
     'ha': 'هَوُسَ',
     'haw': 'Hawaiian',
     'he': 'עברית',
     'hi': 'हिन्दी',
     'hne': 'Chhattisgarhi',
     'ho': 'Hiri Motu',
     'hr': 'Hrvatski',
     'hsb': 'Upper Sorbian',
     'ht': 'Kreyòl ayisyen',
     'hu': 'Magyar',
     'hy': 'Հայերեն',
     'hz': 'Otjiherero',
     'ia': 'Interlingua',
     'id': 'Indonesian',
     'ie': 'Interlingue',
     'ig': 'Asụsụ Igbo',
     'ii': 'ꆈꌠ꒿',
     'ik': 'Iñupiaq',
     'ilo': 'Ilokano',
     'ina': 'Interlingua',
     'io': 'Ido',
     'is': 'Íslenska',
     'it': 'Italiano',
     'iu': 'ᐃᓄᒃᑎᑐᑦ',
     'iw': 'עברית',
     'ja': '日本語',
     'jv': 'basa Jawa',
     'ka': 'ქართული',
     'kab': 'Tazwawt',
     'kac': 'Jingpho',
     'kg': 'Kikongo',
     'ki': 'Gĩkũyũ',
     'kj': 'Kuanyama',
     'kk': 'қазақ тілі',
     'kl': 'kalaallisut',
     'km': 'ខ្មែរ',
     'kn': 'ಕನ್ನಡ',
     'ko': '한국어',
     'kok': 'Konkani',
     'kr': 'Kanuri',
     'ks': 'कश्मीरी',
     'ku': 'Kurdî',
     'kv': 'коми кыв',
     'kw': 'Kernewek',
     'ky': 'Кыргызча',
     'la': 'latine',
     'lb': 'Lëtzebuergesch',
     'lg': 'Luganda',
     'li': 'Limburgs',
     'ln': 'Lingála',
     'lo': 'ພາສາລາວ',
     'lt': 'lietuvių kalba',
     'lu': 'Tshiluba',
     'lv': 'latviešu valoda',
     'mai': 'Maithili',
     'mg': 'fiteny malagasy',
     'mh': 'Kajin M̧ajeļ',
     'mhr': 'Eastern Mari',
     'mi': 'te reo Māori',
     'mk': 'македонски јазик',
     'ml': 'മലയാളം',
     'mn': 'монгол',
     'mni': 'Manipuri',
     'mr': 'मराठी',
     'ms': 'بهاس ملايو',
     'mt': 'Malti',
     'my': 'ဗမာစာ',
     'na': 'Ekakairũ Naoero',
     'nb': 'Bokmål',
     'nd': 'isiNdebele',
     'nds': 'Plattdüütsch',
     'ne': 'नेपाली',
     'ng': 'Owambo',
     'nl': 'Nederlands',
     'nn': 'Norsk nynorsk',
     'no': 'Norsk',
     'nr': 'isiNdebele',
     'nso': 'Pedi',
     'nv': 'Diné bizaad',
     'ny': 'chiCheŵa',
     'oc': 'occitan',
     'oj': 'ᐊᓂᔑᓈᐯᒧᐎᓐ',
     'om': 'Afaan Oromoo',
     'or': 'ଓଡ଼ିଆ',
     'os': 'ирон æвзаг',
     'pa': 'ਪੰਜਾਬੀ',
     'pap': 'Papiamentu',
     'pau': 'a tekoi er a Belau',
     'pi': 'पाऴि',
     'pl': 'polski',
     'ps': 'پښتو',
     'pt': 'Português',
     'pt_BR': 'Português do Brasil',
     'qu': 'Runa Simi',
     'rm': 'rumantsch grischun',
     'rn': 'Ikirundi',
     'ro': 'română',
     'ru': 'Pусский',
     'rw': 'Ikinyarwanda',
     'sa': 'संस्कृतम्',
     'sat': 'ᱥᱟᱱᱛᱟᱲᱤ',
     'sc': 'sardu',
     'sd': 'सिन्धी',
     'se': 'Davvisámegiella',
     'sg': 'yângâ tî sängö',
     'shn': 'Shan',
     'si': 'සිංහල',
     'sk': 'slovenčina',
     'sl': 'slovenščina',
     'sm': 'gagana faa Samoa',
     'sn': 'chiShona',
     'so': 'Soomaaliga',
     'sq': 'Shqip',
     'sr': 'Српски',
     'ss': 'SiSwati',
     'st': 'Sesotho',
     'su': 'Basa Sunda',
     'sv': 'svenska',
     'sw': 'Kiswahili',
     'ta': 'தமிழ்',
     'te': 'తెలుగు',
     'tet': 'Tetum',
     'tg': 'тоҷикӣ',
     'th': 'ไทย',
     'ti': 'ትግርኛ',
     'tig': 'Tigre',
     'tk': 'Türkmen',
     'tl': 'ᜏᜒᜃᜅ᜔ ᜆᜄᜎᜓᜄ᜔',
     'tn': 'Setswana',
     'to': 'faka Tonga',
     'tr': 'Türkçe',
     'ts': 'Xitsonga',
     'tt': 'татар теле',
     'tw': 'Twi',
     'ty': 'Reo Tahiti',
     'ug': 'Uyghur',
     'uk': 'Українська',
     'ur': 'اردو',
     'uz': 'Ўзбек',
     've': 'Tshivenḓa',
     'vi': 'Tiếng Việt',
     'vo': 'Volapük',
     'wa': 'walon',
     'wae': 'Walser',
     'wal': 'Wolaytta',
     'wo': 'Wollof',
     'xh': 'isiXhosa',
     'yi': 'ייִדיש',
     'yo': 'Yorùbá',
     'za': 'Saɯ cueŋƅ',
     'zh': '中文',
     'zh_CN': '中文',
     'zh_TW': '中文',
     'zu': 'isiZulu'}


def get_supported_language_codes():
    """Return list of supported languages as language codes

    Supported means a translation may be available.
    """
    supported_langs = []
    # Use local import to avoid circular import.
    from bleachbit import locale_dir
    # The locale_dir may not exist, especially on Windows.
    if not os.path.isdir(locale_dir):
        return ['en_US', 'en']
    lang_codes = sorted(set(os.listdir(locale_dir) + ['en_US', 'en']))
    for lang in lang_codes:
        if lang in ('en', 'en_US'):
            supported_langs.append(lang)
            continue
        if os.path.isdir(os.path.join(locale_dir, lang)):
            try:
                translation = gettext.translation(
                    'bleachbit', locale_dir, languages=[lang])
                if translation:
                    supported_langs.append(lang)
            except FileNotFoundError:
                pass
    return supported_langs


def get_supported_language_code_name_dict():
    """Return dictionary of supported languages as language codes and names

    Supported means a translation is available.
    """
    supported_langs = {}
    for lang in get_supported_language_codes():
        supported_langs[lang] = native_locale_names.get(lang, None)
    return supported_langs


def get_active_language_code():
    """Return the language ID to use for translations

    The language ID is a code like: en, en_US, nds, C

    There may be an underscore or no underscore. The first part may
    contain two or three letters.

    There will not be a dot like `en_US.UTF-8`.
    """
    try:
        from bleachbit.Options import options
    except ImportError:
        logger.error("Failed to get language options")
    else:
        if not options.get('auto_detect_lang') and options.has_option('forced_language') and options.get('forced_language'):
            return options.get('forced_language')
    import locale
    # locale.getdefaultlocale() will be removed in Python 3.15, so
    # use getlocale() instead.
    # However, on Windows, getlocale() may return values like
    # 'English_United States' instead of RFC1766 codes.
    if os.name == 'nt':
        import ctypes
        kernel32 = ctypes.windll.kernel32
        lcid = kernel32.GetUserDefaultLCID()
        # Convert Windows LCID (e.g., 1033) to RFC1766 (e.g., en-US).
        user_locale = locale.windows_locale.get(lcid, '')
    else:
        user_locale = locale.getlocale()[0]

    if not user_locale:
        user_locale = 'C'
        logger.warning("no default locale found.  Assuming '%s'", user_locale)

    if '.' in user_locale:
        # This should never happen.
        logger.warning('locale contains a dot: %s', user_locale)
        user_locale = user_locale.split('.')[0]

    assert isinstance(user_locale, str)
    assert len(
        user_locale) >= 2 or user_locale == 'C', f"user_locale: {user_locale}"

    return user_locale


def setup_translation():
    """Do a one-time setup of translations"""
    global attempted_setup_translation, t
    attempted_setup_translation = True
    # Use local import to avoid circular import.
    from bleachbit import locale_dir
    user_locale = get_active_language_code()
    logger.debug(f"user_locale: {user_locale}, locale_dir: {locale_dir}")
    assert isinstance(user_locale, str)
    assert isinstance(locale_dir, str), f"locale_dir: {locale_dir}"
    if 'win32' == sys.platform and user_locale:
        os.environ['LANG'] = user_locale
    text_domain = 'bleachbit'
    try:
        t = gettext.translation(
            domain=text_domain, localedir=locale_dir, languages=[user_locale], fallback=True)
    except FileNotFoundError as e:
        logger.error(
            "Error in setup_translation() with language code %s: %s", user_locale, e)
        t = None
        return
    import locale
    if hasattr(locale, 'bindtextdomain'):
        locale.bindtextdomain(text_domain, locale_dir)
        locale.textdomain(text_domain)
    elif 'nt' == os.name:
        from bleachbit.Windows import load_i18n_dll
        libintl = load_i18n_dll()
        if not libintl:
            logger.error(
                'The internationalization library is not available.')
        assert isinstance(text_domain, str)
        encoded_domain = text_domain.encode('utf-8')
        # wbindtextdomain(char, wchar): first parameter is encoded
        if hasattr(libintl, 'libintl_wbindtextdomain'):
            libintl.libintl_wbindtextdomain(encoded_domain, locale_dir)
            libintl.textdomain(encoded_domain)
            libintl.bind_textdomain_codeset(encoded_domain, b'UTF-8')
        else:
            logger.error(
                'The function wbindtextdomain() is not available.')

        # Log for debugging
        logger.debug(
            f"Windows translation domain set to: {text_domain}, dir: {locale_dir}")
    else:
        logger.error('The function bindtextdomain() is not available.')

    # locale.setlocale() on Linux will throw an exception if the locale is not
    # available, so find the best matching locale. When set, Gtk.Builder is
    # translated.
    # On Windows, locale.setlocale() accepts any values without raising an exception.
    if os.name == 'posix':
        from bleachbit.Unix import find_best_locale
        setlocale_local = find_best_locale(user_locale)
        if 'C' == setlocale_local and not user_locale == 'C':
            logger.warning(
                'locale %s is not available. You may wish to run sudo local-gen to generate it.', user_locale)
        try:
            locale.setlocale(locale.LC_ALL, setlocale_local)
        except locale.Error as e:
            logger.error('locale.setlocale(%s): %s:', setlocale_local, e)


def get_text(str):
    """Return translated string

    The name has an underscore to avoid conflicting with gettext module.
    """
    global attempted_setup_translation, t
    if not attempted_setup_translation:
        setup_translation()
    if not t:
        return str
    return t.gettext(str)


def nget_text(singular, plural, n):
    """Return translated string with plural variant"""
    global t
    if not t:
        if 1 == n:
            return singular
        return plural
    return t.ngettext(singular, plural, n)


def pget_text(msgctxt, msgid):
    """Return translated string with context

    Example context is button
    """
    global t
    if not t:
        return msgid
    return t.pgettext(msgctxt, msgid)


attempted_setup_translation = False
t = None



# markovify
# Copyright (c) 2015, Jeremy Singer-Vine
# Origin: https://github.com/jsvine/markovify
# MIT License: https://github.com/jsvine/markovify/blob/master/LICENSE.txt

import random
import operator
import bisect
import json

# Python3 compatibility
try: # pragma: no cover
    basestring
except NameError: # pragma: no cover
    basestring = str

BEGIN = "___BEGIN__"
END = "___END__"

def accumulate(iterable, func=operator.add):
    """
    Cumulative calculations. (Summation, by default.)
    Via: https://docs.python.org/3/library/itertools.html#itertools.accumulate
    """
    it = iter(iterable)
    total = next(it)
    yield total
    for element in it:
        total = func(total, element)
        yield total

class Chain(object):
    """
    A Markov chain representing processes that have both beginnings and ends.
    For example: Sentences.
    """
    def __init__(self, corpus, state_size, model=None):
        """
        `corpus`: A list of lists, where each outer list is a "run"
        of the process (e.g., a single sentence), and each inner list
        contains the steps (e.g., words) in the run. If you want to simulate
        an infinite process, you can come very close by passing just one, very
        long run.

        `state_size`: An integer indicating the number of items the model
        uses to represent its state. For text generation, 2 or 3 are typical.
        """
        self.state_size = state_size
        self.model = model or self.build(corpus, self.state_size)
        self.precompute_begin_state()

    def build(self, corpus, state_size):
        """
        Build a Python representation of the Markov model. Returns a dict
        of dicts where the keys of the outer dict represent all possible states,
        and point to the inner dicts. The inner dicts represent all possibilities
        for the "next" item in the chain, along with the count of times it
        appears.
        """

        # Using a DefaultDict here would be a lot more convenient, however the memory
        # usage is far higher.
        model = {}

        for run in corpus:
            items = ([ BEGIN ] * state_size) + run + [ END ]
            for i in range(len(run) + 1):
                state = tuple(items[i:i+state_size])
                follow = items[i+state_size]
                if state not in model:
                    model[state] = {}

                if follow not in model[state]:
                    model[state][follow] = 0

                model[state][follow] += 1
        return model

    def precompute_begin_state(self):
        """
        Caches the summation calculation and available choices for BEGIN * state_size.
        Significantly speeds up chain generation on large corpuses. Thanks, @schollz!
        """
        begin_state = tuple([ BEGIN ] * self.state_size)
        choices, weights = zip(*self.model[begin_state].items())
        cumdist = list(accumulate(weights))
        self.begin_cumdist = cumdist
        self.begin_choices = choices

    def move(self, state):
        """
        Given a state, choose the next item at random.
        """
        if state == tuple([ BEGIN ] * self.state_size):
            choices = self.begin_choices
            cumdist = self.begin_cumdist
        else:
            choices, weights = zip(*self.model[state].items())
            cumdist = list(accumulate(weights))
        r = random.random() * cumdist[-1]
        selection = choices[bisect.bisect(cumdist, r)]
        return selection

    def gen(self, init_state=None):
        """
        Starting either with a naive BEGIN state, or the provided `init_state`
        (as a tuple), return a generator that will yield successive items
        until the chain reaches the END state.
        """
        state = init_state or (BEGIN,) * self.state_size
        while True:
            next_word = self.move(state)
            if next_word == END: break
            yield next_word
            state = tuple(state[1:]) + (next_word,)

    def walk(self, init_state=None):
        """
        Return a list representing a single run of the Markov model, either
        starting with a naive BEGIN state, or the provided `init_state`
        (as a tuple).
        """
        return list(self.gen(init_state))

    def to_json(self):
        """
        Dump the model as a JSON object, for loading later.
        """
        return json.dumps(list(self.model.items()))

    @classmethod
    def from_json(cls, json_thing):
        """
        Given a JSON object or JSON string that was created by `self.to_json`,
        return the corresponding markovify.Chain.
        """

        if isinstance(json_thing, basestring):
            obj = json.loads(json_thing)
        else:
            obj = json_thing

        if isinstance(obj, list):
            rehydrated = {tuple(item[0]): item[1] for item in obj}
        elif isinstance(obj, dict):
            rehydrated = obj
        else:
            raise ValueError("Object should be dict or list")

        state_size = len(list(rehydrated.keys())[0])

        inst = cls(None, state_size, rehydrated)
        return inst



# -*- coding: utf-8 -*- 
# markovify
# Copyright (c) 2015, Jeremy Singer-Vine
# Origin: https://github.com/jsvine/markovify
# MIT License: https://github.com/jsvine/markovify/blob/master/LICENSE.txt

import re

ascii_lowercase = "abcdefghijklmnopqrstuvwxyz"
ascii_uppercase = ascii_lowercase.upper()

# States w/ with thanks to https://github.com/unitedstates/python-us
# Titles w/ thanks to https://github.com/nytimes/emphasis and @donohoe
abbr_capped = "|".join([
    "ala|ariz|ark|calif|colo|conn|del|fla|ga|ill|ind|kan|ky|la|md|mass|mich|minn|miss|mo|mont|neb|nev|okla|ore|pa|tenn|vt|va|wash|wis|wyo", # States
    "u.s",
    "mr|ms|mrs|msr|dr|gov|pres|sen|sens|rep|reps|prof|gen|messrs|col|sr|jf|sgt|mgr|fr|rev|jr|snr|atty|supt", # Titles
    "ave|blvd|st|rd|hwy", # Streets
    "jan|feb|mar|apr|jun|jul|aug|sep|sept|oct|nov|dec", # Months
    "|".join(ascii_lowercase) # Initials
]).split("|")

abbr_lowercase = "etc|v|vs|viz|al|pct"

exceptions = "U.S.|U.N.|E.U.|F.B.I.|C.I.A.".split("|")

def is_abbreviation(dotted_word):
    clipped = dotted_word[:-1]
    if clipped[0] in ascii_uppercase:
        if clipped.lower() in abbr_capped: return True
        else: return False
    else:
        if clipped in abbr_lowercase: return True
        else: return False

def is_sentence_ender(word):
    if word in exceptions: return False
    if word[-1] in [ "?", "!" ]:
        return True
    if len(re.sub(r"[^A-Z]", "", word)) > 1:
        return True
    if word[-1] == "." and (not is_abbreviation(word)):
        return True
    return False

def split_into_sentences(text):
    potential_end_pat = re.compile(r"".join([
        r"([\w\.'’&\]\)]+[\.\?!])", # A word that ends with punctuation
        r"([‘’“”'\"\)\]]*)", # Followed by optional quote/parens/etc
        r"(\s+(?![a-z\-–—]))", # Followed by whitespace + non-(lowercase or dash)
        ]), re.U)
    dot_iter = re.finditer(potential_end_pat, text)
    end_indices = [ (x.start() + len(x.group(1)) + len(x.group(2)))
        for x in dot_iter
        if is_sentence_ender(x.group(1)) ]
    spans = zip([None] + end_indices, end_indices + [None])
    sentences = [ text[start:end].strip() for start, end in spans ]
    return sentences



# vim: ts=4:sw=4:expandtab
# -*- coding: UTF-8 -*-

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Integration specific to Unix-like operating systems
"""

import configparser
import glob
import logging
import os
import platform
import re
import shlex
import subprocess
import sys

import bleachbit
from bleachbit import FileUtilities, General
from bleachbit.General import get_real_uid, get_real_username
from bleachbit.Language import get_text as _, native_locale_names

logger = logging.getLogger(__name__)

try:
    Pattern = re.Pattern
except AttributeError:
    Pattern = re._pattern_type


JOURNALD_REGEX = r'^Vacuuming done, freed ([\d.]+[BKMGT]?) of archived journals (on disk|from [\w/]+).$'


class LocaleCleanerPath:
    """This represents a path with either a specific folder name or a folder name pattern.
    It also may contain several compiled regex patterns for localization items (folders or files)
    and additional LocaleCleanerPaths that get traversed when asked to supply a list of localization
    items"""

    def __init__(self, location):
        if location is None:
            raise RuntimeError("location is none")
        self.pattern = location
        self.children = []

    def add_child(self, child):
        """Adds a child LocaleCleanerPath"""
        self.children.append(child)
        return child

    def add_path_filter(self, pre, post):
        r"""Adds a filter consisting of a prefix and a postfix
        (e.g. 'foobar_' and '\.qm' to match 'foobar_en_US.utf-8.qm)"""
        try:
            regex = re.compile('^' + pre + Locales.localepattern + post + '$')
        except Exception as errormsg:
            raise RuntimeError(
                f"Malformed regex '{pre}' or '{post}': {errormsg}") from errormsg
        self.add_child(regex)

    def get_subpaths(self, basepath):
        """Returns direct subpaths for this object, i.e. either the named subfolder or all
        subfolders matching the pattern"""
        if isinstance(self.pattern, Pattern):
            return (os.path.join(basepath, p) for p in os.listdir(basepath)
                    if self.pattern.match(p) and os.path.isdir(os.path.join(basepath, p)))
        path = os.path.join(basepath, self.pattern)
        return [path] if os.path.isdir(path) else []

    def get_localizations(self, basepath):
        """Returns all localization items for this object and all descendant objects"""
        for path in self.get_subpaths(basepath):
            for child in self.children:
                if isinstance(child, LocaleCleanerPath):
                    yield from child.get_localizations(path)
                elif isinstance(child, Pattern):
                    for element in os.listdir(path):
                        match = child.match(element)
                        if match is not None:
                            yield (match.group('locale'),
                                   match.group('specifier'),
                                   os.path.join(path, element))


class Locales:
    """Find languages and localization files"""

    # The regular expression to match locale strings and extract the langcode.
    # See test_locale_regex() in tests/TestUnix.py for examples
    # This doesn't match all possible valid locale strings to avoid
    # matching filenames you might want to keep, e.g. the regex
    # to match jp.eucJP might also match jp.importantfileextension
    localepattern =\
        r'(?P<locale>[a-z]{2,3})' \
        r'(?P<specifier>[_-][A-Z]{2,4})?(?:\.[\w]+[\d-]+|@\w+)?' \
        r'(?P<encoding>[.-_](?:(?:ISO|iso|UTF|utf|us-ascii)[\d-]+|(?:euc|EUC)[A-Z]+))?'

    def __init__(self):
        self._paths = LocaleCleanerPath(location='/')

    def add_xml(self, xml_node, parent=None):
        """Parses the xml data and adds nodes to the LocaleCleanerPath-tree"""

        if parent is None:
            parent = self._paths
        if xml_node.ELEMENT_NODE != xml_node.nodeType:
            return

        # if a pattern is supplied, we recurse into all matching subdirectories
        if 'regexfilter' == xml_node.nodeName:
            pre = xml_node.getAttribute('prefix') or ''
            post = xml_node.getAttribute('postfix') or ''
            parent.add_path_filter(pre, post)
        elif 'path' == xml_node.nodeName:
            if xml_node.hasAttribute('directoryregex'):
                pattern = xml_node.getAttribute('directoryregex')
                if '/' in pattern:
                    raise RuntimeError(
                        'directoryregex may not contain slashes.')
                pattern = re.compile(pattern)
                parent = parent.add_child(LocaleCleanerPath(pattern))

            # a combination of directoryregex and filter could be too much
            else:
                if xml_node.hasAttribute("location"):
                    # if there's a filter attribute, it should apply to this path
                    parent = parent.add_child(LocaleCleanerPath(
                        xml_node.getAttribute('location')))

                if xml_node.hasAttribute('filter'):
                    userfilter = xml_node.getAttribute('filter')
                    if 1 != userfilter.count('*'):
                        raise RuntimeError(
                            f"Filter string '{userfilter}' must contain the placeholder * exactly once")

                    # we can't use re.escape, because it escapes too much
                    (pre, post) = (re.sub(r'([\[\]()^$.])', r'\\\1', p)
                                   for p in userfilter.split('*'))
                    parent.add_path_filter(pre, post)
        else:
            raise RuntimeError(
                f"Invalid node '{xml_node.nodeName}', expected '<path>' or '<regexfilter>'")

        # handle child nodes
        for child_xml in xml_node.childNodes:
            self.add_xml(child_xml, parent)

    def localization_paths(self, locales_to_keep):
        """Returns all localization items matching the previously added xml configuration"""
        purgeable_locales = get_purgeable_locales(locales_to_keep)

        for (locale, specifier, path) in self._paths.get_localizations('/'):
            specific = locale + (specifier or '')
            if specific in purgeable_locales or \
                    (locale in purgeable_locales and specific not in locales_to_keep):
                yield path


def _is_broken_xdg_desktop_application(config, desktop_pathname):
    """Returns whether application .desktop file is critically broken

    This function tests only .desktop files with Type=Application.
    """
    if not config.has_option('Desktop Entry', 'Exec'):
        logger.info(
            "is_broken_xdg_menu: missing required option 'Exec' in '%s'", desktop_pathname)
        return True
    exe = config.get('Desktop Entry', 'Exec').split(" ")[0]
    if not FileUtilities.exe_exists(exe):
        logger.info(
            "is_broken_xdg_menu: executable '%s' does not exist in '%s'", exe, desktop_pathname)
        return True
    if 'env' == exe:
        # Wine v1.0 creates .desktop files like this
        # Exec=env WINEPREFIX="/home/z/.wine" wine "C:\\Program
        # Files\\foo\\foo.exe"
        exec_val = config.get('Desktop Entry', 'Exec')
        try:
            execs = shlex.split(exec_val)
        except ValueError as e:
            logger.info(
                "is_broken_xdg_menu: error splitting 'Exec' key '%s' in '%s'", e, desktop_pathname)
            return True
        wineprefix = None
        del execs[0]
        while True:
            if execs[0].find("=") < 0:
                break
            (name, value) = execs[0].split("=")
            if name == 'WINEPREFIX':
                wineprefix = value
            del execs[0]
        if not FileUtilities.exe_exists(execs[0]):
            logger.info(
                "is_broken_xdg_menu: executable '%s' does not exist in '%s'", execs[0], desktop_pathname)
            return True
        # check the Windows executable exists
        if wineprefix:
            windows_exe = wine_to_linux_path(wineprefix, execs[1])
            if not os.path.exists(windows_exe):
                logger.info("is_broken_xdg_menu: Windows executable '%s' does not exist in '%s'",
                            windows_exe, desktop_pathname)
                return True
    return False


def find_available_locales():
    """Returns a list of available locales using locale -a"""
    rc, stdout, stderr = General.run_external(['locale', '-a'])
    if rc == 0:
        return stdout.strip().split('\n')

    logger.warning("Failed to get available locales: %s", stderr)
    return []


def find_best_locale(user_locale):
    """Find closest match to available locales"""
    assert isinstance(user_locale, str)
    if not user_locale:
        return 'C'
    if user_locale in ('C', 'C.utf8', 'POSIX'):
        return user_locale
    available_locales = find_available_locales()

    # If requesting a language like 'es' and current locale is compatible
    # like 'es_MX', then return that.
    # Import here for mock patch.
    import locale  # pylint: disable=import-outside-toplevel
    current_locale = locale.getlocale()[0]
    if current_locale and current_locale.startswith(user_locale.split('.')[0]):
        return '.'.join(locale.getlocale())

    # Check for exact match.
    if user_locale in available_locales:
        return user_locale

    # Next, match like 'en' to 'en_US.utf8' (if available) because
    # of preference for UTF-8.
    for avail_locale in available_locales:
        if avail_locale.startswith(user_locale) and avail_locale.endswith('.utf8'):
            return avail_locale

    # Next, match like 'en' to 'en_US' or 'en_US.iso88591'.
    for avail_locale in available_locales:
        if avail_locale.startswith(user_locale):
            return avail_locale

    return 'C'


def get_distribution_name_version_platform_freedesktop():
    """Returns the name and version of the distribution using
    platform.freedesktop_os_release()

    Example return value: 'ubuntu 24.10' or None

    Python 3.10 added platform.freedesktop_os_release().
    """
    if hasattr(platform, 'freedesktop_os_release'):
        try:
            release = platform.freedesktop_os_release()
        except FileNotFoundError:
            return None
        dist_id = release.get('ID')
        dist_version_id = release.get('VERSION_ID')
        if dist_id and dist_version_id:
            return f"{dist_id} {dist_version_id}"
    return None


def get_distribution_name_version_distro():
    """Returns the name and version of the distribution using the distro
    package

    Example return value: 'ubuntu 24.10' or None

    distro is a third-party package recommended here:
    https://docs.python.org/3.7/library/platform.html
    """
    try:
        # Import here in case of ImportError.
        import distro  # pylint: disable=import-outside-toplevel
        # example 'ubuntu 24.10'
        return distro.id() + ' ' + distro.version()
    except ImportError:
        return None


def get_distribution_name_version_os_release():
    """Returns the name and version of the distribution using /etc/os-release

    Example return value: 'ubuntu 24.10' or None
    """
    if not os.path.exists('/etc/os-release'):
        return None
    try:
        with open('/etc/os-release', 'r', encoding='utf-8') as f:
            os_release = {}
            for line in f:
                if '=' in line:
                    key, value = line.rstrip().split('=', 1)
                    os_release[key] = value.strip('"\'')
    except Exception as e:
        logger.debug("Error reading /etc/os-release: %s", e)
        return None
    if 'ID' in os_release and 'VERSION_ID' in os_release:
        dist_name = os_release['ID']
        return f"{dist_name} {os_release['VERSION_ID']}"
    return None


def get_distribution_name_version():
    """Returns the name and version of the distribution

    Depending on system capabilities, return value may be:
    * 'ubuntu 24.10'
    * 'Linux 6.12.3 (unknown distribution)'
    * 'Linux (unknown version and distribution)'

    Python 3.7 had platform.linux_distribution(), but it
    was removed in Python 3.8.
    """
    ret = get_distribution_name_version_platform_freedesktop()
    if ret:
        return ret
    ret = get_distribution_name_version_distro()
    if ret:
        return ret
    ret = get_distribution_name_version_os_release()
    if ret:
        return ret
    try:
        linux_version = platform.release()
        # example '6.12.3-061203-generic'
        linux_version = linux_version.split('-')[0]
        return f"Linux {linux_version} (unknown distribution)"
    except Exception as e1:
        logger.debug("Error calling platform.release(): %s", e1)
        try:
            linux_version = os.uname().release
            # example '6.12.3-061203-generic'
            linux_version = linux_version.split('-')[0]
            return f"Linux {linux_version} (unknown distribution)"
        except Exception as e2:
            logger.debug("Error calling os.uname(): %s", e2)
    return "Linux (unknown version and distribution)"


def get_purgeable_locales(locales_to_keep):
    """Returns all locales to be purged"""
    if not locales_to_keep:
        raise RuntimeError('Found no locales to keep')

    assert isinstance(locales_to_keep, list)

    # Start with all locales as potentially purgeable
    purgeable_locales = set(native_locale_names.keys())

    # Remove the locales we want to keep
    for keep in locales_to_keep:
        purgeable_locales.discard(keep)
        # If keeping a variant (e.g. 'en_US'), also keep the base locale (e.g. 'en')
        if '_' in keep:
            purgeable_locales.discard(keep[:keep.find('_')])
        # If keeping a base locale (e.g. 'en'), also keep all its variants (e.g. 'en_US')
        if '_' not in keep:
            purgeable_locales = {locale for locale in purgeable_locales
                                 if not locale.startswith(keep + '_')}

    return frozenset(purgeable_locales)


def is_unregistered_mime(mimetype):
    """Returns True if the MIME type is known to be unregistered. If
    registered or unknown, conservatively returns False."""
    try:
        from gi.repository import Gio  # pylint: disable=import-outside-toplevel
        if 0 == len(Gio.app_info_get_all_for_type(mimetype)):
            return True
    except ImportError:
        logger.warning(
            'error calling gio.app_info_get_all_for_type(%s)', mimetype)
    return False


def is_broken_xdg_desktop(pathname):
    """Returns whether the given XDG .desktop file is critically broken.
    Reference: http://standards.freedesktop.org/desktop-entry-spec/latest/"""
    config = bleachbit.RawConfigParser()
    try:
        config.read(pathname)
    except UnicodeDecodeError:
        logger.info(
            "is_broken_xdg_menu: cannot decode file: '%s'", pathname)
        return True
    except (configparser.Error) as e:
        logger.info(
            "is_broken_xdg_menu: %s: '%s'", e, pathname)
        return True
    if not config.has_section('Desktop Entry'):
        logger.info(
            "is_broken_xdg_menu: missing required section 'Desktop Entry': '%s'", pathname)
        return True
    if not config.has_option('Desktop Entry', 'Type'):
        logger.info(
            "is_broken_xdg_menu: missing required option 'Type': '%s'", pathname)
        return True
    file_type = config.get('Desktop Entry', 'Type').strip().lower()
    if 'link' == file_type:
        if not config.has_option('Desktop Entry', 'URL') and \
                not config.has_option('Desktop Entry', 'URL[$e]'):
            logger.info(
                "is_broken_xdg_menu: missing required option 'URL': '%s'", pathname)
            return True
        return False
    if 'mimetype' == file_type:
        if not config.has_option('Desktop Entry', 'MimeType'):
            logger.info(
                "is_broken_xdg_menu: missing required option 'MimeType': '%s'", pathname)
            return True
        mimetype = config.get('Desktop Entry', 'MimeType').strip().lower()
        if is_unregistered_mime(mimetype):
            logger.info(
                "is_broken_xdg_menu: MimeType '%s' not registered '%s'", mimetype, pathname)
            return True
        return False
    if 'application' != file_type:
        logger.warning("unhandled type '%s': file '%s'", file_type, pathname)
        return False
    if _is_broken_xdg_desktop_application(config, pathname):
        return True
    return False


def is_process_running_ps_aux(exename, require_same_user):
    """Check whether exename is running by calling 'ps aux -c'

    exename: name of the executable
    require_same_user: if True, ignore processes run by other users

    When running under sudo, this uses the non-root username.
    """
    ps_out = subprocess.check_output(["ps", "aux", "-c"],
                                     universal_newlines=True)
    first_line = ps_out.split('\n', maxsplit=1)[0].strip()
    if "USER" not in first_line or "COMMAND" not in first_line:
        raise RuntimeError("Unexpected ps header format")

    for line in ps_out.split("\n")[1:]:
        parts = line.split()
        if len(parts) < 11:
            continue
        process_user = parts[0]
        process_cmd = parts[10]
        if process_cmd != exename:
            continue
        if not require_same_user or process_user == get_real_username():
            return True
    return False


def is_process_running_linux(exename, require_same_user):
    """Check whether exename is running

    The exename is checked two different ways.

    When running under sudo, this uses the non-root user ID.
    """
    for filename in glob.iglob("/proc/*/exe"):
        does_exe_match = False
        try:
            target = os.path.realpath(filename)
        except TypeError:
            # happens, for example, when link points to
            # '/etc/password\x00 (deleted)'
            pass
        except OSError:
            # 13 = permission denied
            pass
        else:
            # Google Chrome 74 on Ubuntu 19.04 shows up as
            # /opt/google/chrome/chrome (deleted)
            found_exename = os.path.basename(target).replace(' (deleted)', '')
            does_exe_match = exename == found_exename

        if not does_exe_match:
            with open(os.path.join(os.path.dirname(filename), 'stat'), 'r', encoding='utf-8') as stat_file:
                proc_name = stat_file.read().split()[1].strip('()')
                if proc_name == exename:
                    does_exe_match = True
                else:
                    continue

        if not require_same_user:
            return True

        try:
            uid = os.stat(os.path.dirname(filename)).st_uid
        except OSError:
            # permission denied means not the same user
            continue
        # In case of sudo, use the regular user's ID.
        if uid == get_real_uid():
            return True
    return False


def is_process_running(exename, require_same_user):
    """Check whether exename is running

    exename: name of the executable
    require_same_user: if True, ignore processes run by other users

    """
    if sys.platform == 'linux':
        return is_process_running_linux(exename, require_same_user)
    if sys.platform == 'darwin' or sys.platform.startswith('openbsd') or sys.platform.startswith('freebsd'):
        return is_process_running_ps_aux(exename, require_same_user)
    raise RuntimeError('unsupported platform for is_process_running()')


def rotated_logs():
    """Yield a list of rotated (i.e., old) logs in /var/log/

    See:
    https://bugs.launchpad.net/bleachbit/+bug/367575
    https://github.com/bleachbit/bleachbit/issues/1744
    """
    whitelists = [re.compile(r'/var/log/(removed_)?(packages|scripts)'),
                  re.compile(r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}')]
    positive_re = re.compile(r'(\.(\d+|bz2|gz|xz|old)|\-\d{8}?)')

    for path in bleachbit.FileUtilities.children_in_directory('/var/log'):
        whitelist_match = False
        for whitelist in whitelists:
            if whitelist.search(path) or bleachbit.FileUtilities.whitelisted(path):
                whitelist_match = True
                break
        if whitelist_match:
            continue
        if positive_re.search(path):
            yield path


def wine_to_linux_path(wineprefix, windows_pathname):
    """Return a Linux pathname from an absolute Windows pathname and Wine prefix"""
    drive_letter = windows_pathname[0]
    windows_pathname = windows_pathname.replace(drive_letter + ":",
                                                "drive_" + drive_letter.lower())
    windows_pathname = windows_pathname.replace("\\", "/")
    return os.path.join(wineprefix, windows_pathname)


def run_cleaner_cmd(cmd, args, freed_space_regex=r'[\d.]+[kMGTE]?B?', error_line_regexes=None):
    """Runs a specified command and returns how much space was (reportedly) freed.
    The subprocess shouldn't need any user input and the user should have the
    necessary rights.
    freed_space_regex gets applied to every output line, if the re matches,
    add values captured by the single group in the regex"""
    if not FileUtilities.exe_exists(cmd):
        raise RuntimeError(_('Executable not found: %s') % cmd)
    freed_space_regex = re.compile(freed_space_regex)
    error_line_regexes = [re.compile(regex)
                          for regex in error_line_regexes or []]

    env = {'LC_ALL': 'C', 'PATH': os.getenv('PATH')}
    output = subprocess.check_output([cmd] + args, stderr=subprocess.STDOUT,
                                     universal_newlines=True, env=env)
    freed_space = 0
    for line in output.split('\n'):
        m = freed_space_regex.match(line)
        if m is not None:
            freed_space += FileUtilities.human_to_bytes(m.group(1))
        for error_re in error_line_regexes:
            if error_re.search(line):
                raise RuntimeError('Invalid output from %s: %s' % (cmd, line))

    return freed_space


def journald_clean():
    """Clean the system journals"""
    try:
        return run_cleaner_cmd('journalctl', ['--vacuum-size=1'], JOURNALD_REGEX)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Error calling '{' '.join(e.cmd)}':\n{e.output}") from e


def apt_autoremove():
    """Run 'apt-get autoremove' and return the size (un-rounded, in bytes) of freed space"""

    args = ['--yes', 'autoremove']
    # After this operation, 74.7MB disk space will be freed.
    # After this operation, 44.0 kB disk space will be freed.
    freed_space_regex = r'.*, ([\d.]+ ?[a-zA-Z]{2}) disk space will be freed.'
    try:
        return run_cleaner_cmd('apt-get', args, freed_space_regex, ['^E: '])
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Error calling '{' '.join(e.cmd)}':\n{e.output}") from e


def apt_autoclean():
    """Run 'apt-get autoclean' and return the size (un-rounded, in bytes) of freed space"""
    try:
        return run_cleaner_cmd('apt-get', ['autoclean'], r'^Del .*\[([\d.]+[a-zA-Z]{2})}]', ['^E: '])
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Error calling '{' '.join(e.cmd)}':\n{e.output}") from e


def apt_clean():
    """Run 'apt-get clean' and return the size in bytes of freed space"""
    old_size = get_apt_size()
    try:
        run_cleaner_cmd('apt-get', ['clean'], '^unused regex$', ['^E: '])
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Error calling '{' '.join(e.cmd)}':\n{e.output}") from e
    new_size = get_apt_size()
    return old_size - new_size


def get_apt_size():
    """Return the size of the apt cache (in bytes)"""
    (_rc, stdout, _stderr) = General.run_external(['apt-get', '-s', 'clean'])
    paths = re.findall(r'/[/a-z\.\*]+', stdout)
    return get_globs_size(paths)


def get_globs_size(paths):
    """Get the cumulative size (in bytes) of a list of globs"""
    total_size = 0
    for path in paths:
        for p in glob.iglob(path):
            total_size += FileUtilities.getsize(p)
    return total_size


def yum_clean():
    """Run 'yum clean all' and return size in bytes recovered"""
    if os.path.exists('/var/run/yum.pid'):
        msg = _(
            "%s cannot be cleaned because it is currently running.  Close it, and try again.") % "Yum"
        raise RuntimeError(msg)

    old_size = FileUtilities.getsizedir('/var/cache/yum')
    args = ['--enablerepo=*', 'clean', 'all']
    invalid = ['You need to be root', 'Cannot remove rpmdb file']
    run_cleaner_cmd('yum', args, '^unused regex$', invalid)
    new_size = FileUtilities.getsizedir('/var/cache/yum')
    return old_size - new_size


def dnf_clean():
    """Run 'dnf clean all' and return size in bytes recovered"""
    if os.path.exists('/var/run/dnf.pid'):
        msg = _(
            "%s cannot be cleaned because it is currently running.  Close it, and try again.") % "Dnf"
        raise RuntimeError(msg)

    old_size = FileUtilities.getsizedir('/var/cache/dnf')
    args = ['--enablerepo=*', 'clean', 'all']
    invalid = ['You need to be root', 'Cannot remove rpmdb file']
    run_cleaner_cmd('dnf', args, '^unused regex$', invalid)
    new_size = FileUtilities.getsizedir('/var/cache/dnf')

    return old_size - new_size


units = {"B": 1, "k": 10**3, "M": 10**6, "G": 10**9}


def parse_size(size):
    """Parse the size returned by dnf"""
    number, unit = [string.strip() for string in size.split()]
    return int(float(number) * units[unit])


def dnf_autoremove():
    """Run 'dnf autoremove' and return size in bytes recovered."""
    if os.path.exists('/var/run/dnf.pid'):
        msg = _(
            "%s cannot be cleaned because it is currently running.  Close it, and try again.") % "Dnf"
        raise RuntimeError(msg)
    cmd = ['dnf', '-y', 'autoremove']
    (rc, stdout, stderr) = General.run_external(cmd)
    freed_bytes = 0
    allout = stdout + stderr
    if 'Error: This command has to be run under the root user.' in allout:
        raise RuntimeError('dnf autoremove requires root permissions')
    if rc > 0:
        raise RuntimeError(f'dnf autoremove raised error {rc}: {stderr}')

    cregex = re.compile(r"Freed space: ([\d.]+[\s]+[BkMG])")
    match = cregex.search(allout)
    if match:
        freed_bytes = parse_size(match.group(1))
    logger.debug(
        'dnf_autoremove >> total freed bytes: %s', freed_bytes)
    return freed_bytes


def is_unix_display_protocol_wayland():
    """Return True if the display protocol is Wayland."""
    assert os.name == 'posix'
    if 'XDG_SESSION_TYPE' in os.environ:
        if os.environ['XDG_SESSION_TYPE'] == 'wayland':
            return True
        # If not wayland, then x11, mir, etc.
        return False
    if 'WAYLAND_DISPLAY' in os.environ:
        return True
    # Ubuntu 24.10 showed "ubuntu-xorg".
    # openSUSE Tumbleweed and Fedora 41 showed "gnome".
    # Fedora 41 also showed "plasma".
    if os.environ.get('DESKTOP_SESSION') in ('ubuntu-xorg', 'gnome', 'plasma'):
        return False
    # Wayland (Ubuntu 23.10) sets DISPLAY=:0 like x11, so do not check DISPLAY.
    try:
        (rc, stdout, _stderr) = General.run_external(['loginctl'])
    except FileNotFoundError:
        return False
    if rc != 0:
        logger.warning('logintctl returned rc %s', rc)
        return False
    try:
        session = stdout.split('\n')[1].strip().split(' ')[0]
    except (IndexError, ValueError):
        logger.warning('unexpected output from loginctl: %s', stdout)
        return False
    if not session.isdigit():
        logger.warning('unexpected session loginctl: %s', session)
        return False
    result = General.run_external(
        ['loginctl', 'show-session', session, '-p', 'Type'])
    return 'wayland' in result[1].lower()


def root_is_not_allowed_to_X_session():
    assert os.name == 'posix'
    result = General.run_external(['xhost'], clean_env=False)
    xhost_returned_error = result[0] == 1
    return xhost_returned_error


def is_display_protocol_wayland_and_root_not_allowed():
    try:
        is_wayland = bleachbit.Unix.is_unix_display_protocol_wayland()
    except Exception as e:
        logger.exception(e)
        return False
    return (
        is_wayland and
        os.environ['USER'] == 'root' and
        bleachbit.Unix.root_is_not_allowed_to_X_session()
    )


locales = Locales()



# vim: ts=4:sw=4:expandtab

# BleachBit
# Copyright (C) 2008-2025 Andrew Ziem
# https://www.bleachbit.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Create cleaners from CleanerML (markup language)
"""

# standard library
import logging
import os
import stat
import sys
import xml.dom.minidom

# local import
import bleachbit
from bleachbit.Action import ActionProvider
from bleachbit.FileUtilities import expand_glob_join, listdir
from bleachbit.General import boolstr_to_bool, getText
from bleachbit.Language import get_text as _
from bleachbit import Cleaner

logger = logging.getLogger(__name__)


def default_vars():
    """Return default multi-value variables"""
    ret = {}
    if not os.name == 'nt':
        return ret
    # Expand ProgramFiles to also be ProgramW6432, etc.
    wowvars = (('ProgramFiles', 'ProgramW6432'),
               ('CommonProgramFiles', 'CommonProgramW6432'))
    for v1, v2 in wowvars:
        # Remove None, if variable is not found.
        # Make list unique.
        mylist = list({x for x in (os.getenv(v1), os.getenv(v2)) if x})
        ret[v1] = mylist
    return ret


class CleanerML:

    """Create a cleaner from CleanerML"""

    def __init__(self, pathname, xlate_cb=None):
        """Create cleaner from XML in pathname.

        If xlate_cb is set, use it as a callback for each
        translate-able string.
        """

        self.action = None
        self.cleaner = Cleaner.Cleaner()
        self.option_id = None
        self.option_name = None
        self.option_description = None
        self.option_warning = None
        self.vars = default_vars()
        self.xlate_cb = xlate_cb
        if self.xlate_cb is None:
            self.xlate_mode = False
            self.xlate_cb = lambda x, y=None: None  # do nothing
        else:
            self.xlate_mode = True

        try:
            dom = xml.dom.minidom.parse(pathname)
        except xml.parsers.expat.ExpatError as e:
            logger.error(
                "Error parsing CleanerML file %s with error %s", pathname, e)
            return

        self.handle_cleaner(dom.getElementsByTagName('cleaner')[0])

    def get_cleaner(self):
        """Return the created cleaner"""
        return self.cleaner

    def os_match(self, os_str, platform=sys.platform):
        """Return boolean whether operating system matches

        Keyword arguments:
        os_str -- the required operating system as written in XML
        platform -- used only for unit tests
        """
        # If blank or if in .pot-creation-mode, return true.
        if len(os_str) == 0 or self.xlate_mode:
            return True
        # Otherwise, check platform.
        # Define the current operating system.
        if platform == 'darwin':
            current_os = ('darwin', 'bsd', 'unix')
        elif platform == 'linux':
            current_os = ('linux', 'unix')
        elif platform.startswith('openbsd'):
            current_os = ('bsd', 'openbsd', 'unix')
        elif platform.startswith('netbsd'):
            current_os = ('bsd', 'netbsd', 'unix')
        elif platform.startswith('freebsd'):
            current_os = ('bsd', 'freebsd', 'unix')
        elif platform == 'win32':
            current_os = ('windows',)
        else:
            raise RuntimeError(f'Unknown operating system: {sys.platform}')
        # Compare current OS against required OS.
        return os_str in current_os

    def handle_cleaner(self, cleaner):
        """<cleaner> element"""
        if not self.os_match(cleaner.getAttribute('os')):
            return
        self.cleaner.id = cleaner.getAttribute('id')
        self.handle_cleaner_label(cleaner.getElementsByTagName('label')[0])
        description = cleaner.getElementsByTagName('description')
        if description and description[0].parentNode == cleaner:
            self.handle_cleaner_description(description[0])
        for var in cleaner.getElementsByTagName('var'):
            self.handle_cleaner_var(var)
        for option in cleaner.getElementsByTagName('option'):
            try:
                self.handle_cleaner_option(option)
            except Exception:
                exc_msg = _(
                    "Error in handle_cleaner_option() for cleaner id = {cleaner_id}, option XML={option_xml}")
                logger.exception(exc_msg.format(
                    cleaner_id=self.cleaner.id, option_xml=option.toxml()))
        self.handle_cleaner_running(cleaner.getElementsByTagName('running'))
        self.handle_localizations(
            cleaner.getElementsByTagName('localizations'))

    def handle_cleaner_label(self, label):
        """<label> element under <cleaner>"""
        self.cleaner.name = _(getText(label.childNodes))
        translate = label.getAttribute('translate')
        if translate and boolstr_to_bool(translate):
            self.xlate_cb(self.cleaner.name)

    def handle_cleaner_description(self, description):
        """<description> element under <cleaner>"""
        self.cleaner.description = _(getText(description.childNodes))
        translators = description.getAttribute('translators')
        self.xlate_cb(self.cleaner.description, translators)

    def handle_cleaner_running(self, running_elements):
        """<running> element under <cleaner>"""
        # example: <running type="command">opera</running>
        for running in running_elements:
            if not self.os_match(running.getAttribute('os')):
                continue
            detection_type = running.getAttribute('type')
            value = getText(running.childNodes)
            same_user = running.getAttribute('same_user') or False
            self.cleaner.add_running(detection_type, value, same_user)

    def handle_cleaner_option(self, option):
        """<option> element"""
        self.option_id = option.getAttribute('id')
        self.option_description = None
        self.option_name = None

        self.handle_cleaner_option_label(
            option.getElementsByTagName('label')[0])
        description = option.getElementsByTagName('description')
        self.handle_cleaner_option_description(description[0])
        warning = option.getElementsByTagName('warning')
        if warning:
            self.handle_cleaner_option_warning(warning[0])
            if self.option_warning:
                self.cleaner.set_warning(self.option_id, self.option_warning)

        for action in option.getElementsByTagName('action'):
            self.handle_cleaner_option_action(action)

        self.cleaner.add_option(
            self.option_id, self.option_name, self.option_description)

    def handle_cleaner_option_label(self, label):
        """<label> element under <option>"""
        self.option_name = _(getText(label.childNodes))
        translate = label.getAttribute('translate')
        translators = label.getAttribute('translators')
        if not translate or boolstr_to_bool(translate):
            self.xlate_cb(self.option_name, translators)

    def handle_cleaner_option_description(self, description):
        """<description> element under <option>"""
        self.option_description = _(getText(description.childNodes))
        translators = description.getAttribute('translators')
        self.xlate_cb(self.option_description, translators)

    def handle_cleaner_option_warning(self, warning):
        """<warning> element under <option>"""
        self.option_warning = _(getText(warning.childNodes))
        self.xlate_cb(self.option_warning)

    def handle_cleaner_option_action(self, action_node):
        """<action> element under <option>"""
        if not self.os_match(action_node.getAttribute('os')):
            return
        command = action_node.getAttribute('command')
        provider = None
        for actionplugin in ActionProvider.plugins:
            if actionplugin.action_key == command:
                provider = actionplugin(action_node, self.vars)
        if provider is None:
            raise RuntimeError(f"Invalid command '{command}'")
        self.cleaner.add_action(self.option_id, provider)

    def handle_localizations(self, localization_nodes):
        """<localizations> element under <cleaner>"""
        if not 'posix' == os.name:
            return
        # pylint: disable=import-outside-toplevel
        from bleachbit import Unix
        for localization_node in localization_nodes:
            for child_node in localization_node.childNodes:
                Unix.locales.add_xml(child_node)
        # Add a dummy action so the file isn't reported as unusable
        self.cleaner.add_action('localization', ActionProvider(None))

    def handle_cleaner_var(self, var):
        """Handle one <var> element under <cleaner>.

        Example:

        <var name="basepath">
         <value search="glob">~/.config/f*</value>
         <value>~/.config/foo</value>
         <value>%AppData\foo</value>
         </var>
        """
        var_name = var.getAttribute('name')
        for value_element in var.getElementsByTagName('value'):
            if not self.os_match(value_element.getAttribute('os')):
                continue
            value_str = getText(value_element.childNodes)
            is_glob = value_element.getAttribute('search') == 'glob'
            if is_glob:
                value_list = expand_glob_join(value_str, '')
            else:
                value_list = [value_str, ]
            if var_name in self.vars:
                # append
                self.vars[var_name] = value_list + self.vars[var_name]
            else:
                # initialize
                self.vars[var_name] = value_list


def list_cleanerml_files(local_only=False):
    """List CleanerML files"""
    cleanerdirs = (bleachbit.personal_cleaners_dir, )
    if bleachbit.local_cleaners_dir:
        # If the application is installed, locale_cleaners_dir is None.
        # If portable mode, local_cleaners_dir is under the directory of
        # `bleachbit.py`.
        cleanerdirs += (bleachbit.local_cleaners_dir, )
    if not local_only and bleachbit.system_cleaners_dir:
        cleanerdirs += (bleachbit.system_cleaners_dir, )
    for pathname in listdir(cleanerdirs):
        if not pathname.lower().endswith('.xml'):
            continue
        st = os.stat(pathname)
        if sys.platform != 'win32' and stat.S_IMODE(st[stat.ST_MODE]) & 2:
            # TRANSLATORS: When BleachBit detects the file permissions are
            # insecure, it will not load the cleaner as if it did not exist.
            logger.warning(
                _("Ignoring cleaner because it is world writable: %s"), pathname)
            continue
        yield pathname


def load_cleaners(cb_progress=lambda x: None):
    """Scan for CleanerML and load them"""
    cleanerml_files = list(list_cleanerml_files())
    cleanerml_files.sort()
    if not cleanerml_files:
        logger.debug('No CleanerML files to load.')
        return
    total_files = len(cleanerml_files)
    cb_progress(0.0)
    files_done = 0
    for pathname in cleanerml_files:
        try:
            xmlcleaner = CleanerML(pathname)
        except Exception:
            logger.exception(_("Error reading cleaner: %s"), pathname)
            continue
        cleaner = xmlcleaner.get_cleaner()
        if cleaner.is_usable():
            Cleaner.backends[cleaner.id] = cleaner
        else:
            logger.debug(
                # TRANSLATORS: An action is something like cleaning a specific file.
                # "Not usable" means the whole cleaner will be ignored.
                # The substituted variable is a pathname.
                _("Cleaner is not usable on this OS because it has no actions: %s"), pathname)
        files_done += 1
        cb_progress(1.0 * files_done / total_files)
        yield True


def pot_fragment(msgid, pathname, translators=None):
    """Create a string fragment for generating .pot files"""
    msgid = msgid.replace('"', '\\"')  # escape quotation mark
    if translators:
        translators = f"#. {translators}\n"
    else:
        translators = ""
    ret = f'''{translators}#: {pathname}
msgid "{msgid}"
msgstr ""

'''
    return ret


def create_pot():
    """Create a .pot for translation using gettext"""

    with open('../po/cleanerml.pot', 'w', encoding='utf-8') as f:
        for pathname in listdir('../cleaners'):
            if not pathname.lower().endswith(".xml"):
                continue
            strings = []
            try:
                CleanerML(pathname,
                          lambda newstr, translators=None, current_strings=strings:
                          current_strings.append([newstr, translators]))
            except Exception:
                logger.exception(_("Error reading cleaner: %s"), pathname)
                continue
            for (string, translators) in strings:
                f.write(pot_fragment(string, pathname, translators))
